# pylint: disable=unsubscriptable-object,function-redefined
# Tests for the Indicator objects
from __future__ import annotations

import gc
import json
from inspect import signature

import dask
import numpy as np
import pytest
import xarray as xr

import xclim
from xclim import __version__, atmos
from xclim.core import VARIABLES, MissingVariableError, Quantified
from xclim.core.calendar import select_time
from xclim.core.formatting import (
    AttrFormatter,
    default_formatter,
    merge_attributes,
    parse_doc,
    update_history,
)
from xclim.core.indicator import Daily, Indicator, ResamplingIndicator, registry
from xclim.core.units import convert_units_to, declare_units, units
from xclim.core.utils import InputKind
from xclim.indices import tg_mean
from xclim.testing import list_input_variables


@declare_units(da="[temperature]", thresh="[temperature]")
def uniindtemp_compute(
    da: xr.DataArray,
    thresh: Quantified = "0.0 degC",
    freq: str = "YS",
    method: str = "injected",
):
    """Docstring"""
    out = da - convert_units_to(thresh, da)
    out = out.resample(time=freq).mean()
    out.attrs["units"] = da.units
    return out


uniIndTemp = Daily(
    realm="atmos",
    identifier="tmin",
    module="test",
    cf_attrs=[
        dict(
            var_name="tmin{thresh}",
            units="K",
            long_name="{freq} mean surface temperature with {thresh} threshold.",
            standard_name="{freq} mean temperature",
            cell_methods="time: mean within {freq:noun}",
            another_attr="With a value.",
        )
    ],
    compute=uniindtemp_compute,
    parameters={"method": "injected"},
)


@declare_units(da="[precipitation]")
def uniindpr_compute(da: xr.DataArray, freq: str):
    """Docstring"""
    return da.resample(time=freq).mean(keep_attrs=True)


uniIndPr = Daily(
    realm="atmos",
    identifier="prmax",
    cf_attrs=[dict(units="mm/s")],
    context="hydro",
    module="test",
    compute=uniindpr_compute,
)


@declare_units(da="[temperature]")
def uniclim_compute(da: xr.DataArray, freq="YS", **indexer):
    select = select_time(da, **indexer)
    return select.mean(dim="time", keep_attrs=True).expand_dims("time")


uniClim = ResamplingIndicator(
    src_freq="D",
    realm="atmos",
    identifier="clim",
    cf_attrs=[dict(units="K")],
    module="test",
    compute=uniclim_compute,
)


@declare_units(tas="[temperature]")
def multitemp_compute(tas: xr.DataArray, freq: str):
    return (
        tas.resample(time=freq).min(keep_attrs=True),
        tas.resample(time=freq).max(keep_attrs=True),
    )


multiTemp = Daily(
    realm="atmos",
    identifier="minmaxtemp",
    cf_attrs=[
        dict(
            var_name="tmin",
            units="K",
            standard_name="Min temp",
            description="Grouped computation of tmax and tmin",
        ),
        dict(
            var_name="tmax",
            units="K",
            description="Grouped computation of tmax and tmin",
        ),
    ],
    module="test",
    compute=multitemp_compute,
)


@declare_units(tas="[temperature]", tasmin="[temperature]", tasmax="[temperature]")
def multioptvar_compute(
    tas: xr.DataArray | None = None,
    tasmax: xr.DataArray | None = None,
    tasmin: xr.DataArray | None = None,
):
    if tas is None:
        tasmax = convert_units_to(tasmax, tasmin)
        return ((tasmin + tasmax) / 2).assign_attrs(units=tasmin.units)
    return tas


multiOptVar = Indicator(
    src_freq="D",
    realm="atmos",
    identifier="multiopt",
    cf_attrs=[dict(units="K")],
    module="test",
    compute=multioptvar_compute,
)


def test_attrs(tas_series):
    import datetime as dt

    a = tas_series(np.arange(360.0))
    txm = uniIndTemp(a, thresh="5 degC", freq="YS")
    assert txm.cell_methods == "time: mean time: mean within years"
    assert f"{dt.datetime.now():%Y-%m-%d %H}" in txm.attrs["history"]
    assert (
        "TMIN(da=tas, thresh='5 degC', freq='YS') with options check_missing=any"
        in txm.attrs["history"]
    )
    assert f"xclim version: {__version__}" in txm.attrs["history"]
    assert txm.name == "tmin5 degC"
    assert uniIndTemp.standard_name == "{freq} mean temperature"
    assert uniIndTemp.cf_attrs[0]["another_attr"] == "With a value."

    thresh = xr.DataArray(
        [1],
        dims=("adim",),
        coords={"adim": [1]},
        attrs={"long_name": "A thresh", "units": "degC"},
        name="TT",
    )
    txm = uniIndTemp(a, thresh=thresh, freq="YS")
    assert (
        "TMIN(da=tas, thresh=TT, freq='YS') with options check_missing=any"
        in txm.attrs["history"]
    )
    assert txm.attrs["long_name"].endswith("with <an array> threshold.")


@pytest.mark.parametrize(
    "xcopt,xropt,exp",
    [
        ("xarray", "default", False),
        (True, False, True),
        (False, True, False),
        ("xarray", True, True),
    ],
)
def test_keep_attrs(tasmin_series, tasmax_series, xcopt, xropt, exp):
    tx = tasmax_series(np.arange(360.0))
    tn = tasmin_series(np.arange(360.0))
    tx.attrs.update(something="blabla", bing="bang", foo="bar")
    tn.attrs.update(something="blabla", bing="bong")
    with xclim.set_options(keep_attrs=xcopt):
        with xr.set_options(keep_attrs=xropt):
            tg = multiOptVar(tasmin=tn, tasmax=tx)
    assert (tg.attrs.get("something") == "blabla") is exp
    assert (tg.attrs.get("foo") == "bar") is exp
    assert "bing" not in tg.attrs


def test_as_dataset(tasmax_series, tasmin_series):
    tx = tasmax_series(np.arange(360.0))
    tn = tasmin_series(np.arange(360.0))
    tx.attrs.update(something="blabla", bing="bang", foo="bar")
    tn.attrs.update(something="blabla", bing="bong")
    dsin = xr.Dataset({"tasmax": tx, "tasmin": tn}, attrs={"fou": "barre"})
    with xclim.set_options(keep_attrs=True, as_dataset=True):
        dsout = multiOptVar(ds=dsin)
    assert isinstance(dsout, xr.Dataset)
    assert dsout.attrs["fou"] == "barre"
    assert dsout.multiopt.attrs.get("something") == "blabla"


def test_as_dataset_multi(tas_series):
    tg = tas_series(np.arange(360.0))
    with xclim.set_options(as_dataset=True):
        dsout = multiTemp(tas=tg, freq="YS")
    assert isinstance(dsout, xr.Dataset)
    assert "tmin" in dsout.data_vars
    assert "tmax" in dsout.data_vars


def test_opt_vars(tasmin_series, tasmax_series):
    tn = tasmin_series(np.zeros(365))
    tx = tasmax_series(np.zeros(365))

    multiOptVar(tasmin=tn, tasmax=tx)
    assert multiOptVar.parameters["tasmin"].kind == InputKind.OPTIONAL_VARIABLE


def test_registering():
    assert "test.TMIN" in registry

    # Because this has not been instantiated, it's not in any registry.
    class Test123(registry["test.TMIN"]):
        identifier = "test123"

    assert "test.TEST123" not in registry
    Test123(module="test")
    assert "test.TEST123" in registry

    # Confirm registries live in subclasses.
    class IndicatorNew(Indicator):
        pass

    # Identifier must be given
    with pytest.raises(AttributeError, match="has not been set."):
        IndicatorNew()

    # Realm must be given
    with pytest.raises(AttributeError, match="realm must be given"):
        IndicatorNew(identifier="i2d")

    indnew = IndicatorNew(identifier="i2d", realm="atmos", module="test")
    assert "test.I2D" in registry
    assert registry["test.I2D"].get_instance() is indnew

    del indnew
    gc.collect()
    with pytest.raises(ValueError, match="There is no existing instance"):
        registry["test.I2D"].get_instance()


def test_module():
    """Translations are keyed according to the module where the indicators are defined."""
    assert atmos.tg_mean.__module__.split(".")[2] == "atmos"
    # Virtual module also are stored under xclim.indicators
    assert xclim.indicators.cf.fg.__module__ == "xclim.indicators.cf"  # noqa: F821
    assert (
        xclim.indicators.icclim.GD4.__module__ == "xclim.indicators.icclim"
    )  # noqa: F821


def test_temp_unit_conversion(tas_series):
    a = tas_series(np.arange(365), start="2001-01-01")
    txk = uniIndTemp(a, freq="YS")

    # This is not supposed to work
    uniIndTemp.units = "degC"
    txc = uniIndTemp(a, freq="YS")
    with pytest.raises(AssertionError):
        np.testing.assert_array_almost_equal(txk, txc + 273.15)

    uniIndTemp.cf_attrs[0]["units"] = "degC"
    txc = uniIndTemp(a, freq="YS")
    np.testing.assert_array_almost_equal(txk, txc + 273.15)


def test_temp_diff_unit_conversion(tasmax_series, tasmin_series):
    tx = tasmax_series(np.arange(365) + 1, start="2001-01-01")
    tn = tasmin_series(np.arange(365), start="2001-01-01")
    txC = convert_units_to(tx, "degC")
    tnC = convert_units_to(tn, "degC")

    ind = xclim.atmos.daily_temperature_range.from_dict(
        {"units": "degC"}, "dtr_degC", "test"
    )
    out = ind(tasmax=txC, tasmin=tnC)
    assert out.attrs["units"] == "degC"
    assert out.attrs["units_metadata"] == "temperature: difference"


def test_multiindicator(tas_series):
    tas = tas_series(np.arange(366), start="2000-01-01")
    tmin, tmax = multiTemp(tas, freq="YS")

    assert tmin[0] == tas.min()
    assert tmax[0] == tas.max()
    assert tmin.attrs["standard_name"] == "Min temp"
    assert tmin.attrs["description"] == "Grouped computation of tmax and tmin"
    assert tmax.attrs["description"] == "Grouped computation of tmax and tmin"
    assert multiTemp.units == ["K", "K"]

    # Attrs passed as keywords - together
    ind = Daily(
        realm="atmos",
        identifier="minmaxtemp2",
        cf_attrs=[
            dict(
                var_name="tmin",
                units="K",
                standard_name="Min temp",
                description="Grouped computation of tmax and tmin",
            ),
            dict(
                var_name="tmax",
                units="K",
                description="Grouped computation of tmax and tmin",
            ),
        ],
        compute=multitemp_compute,
    )
    tmin, tmax = ind(tas, freq="YS")
    assert tmin[0] == tas.min()
    assert tmax[0] == tas.max()
    assert tmin.attrs["standard_name"] == "Min temp"
    assert tmin.attrs["description"] == "Grouped computation of tmax and tmin"
    assert tmax.attrs["description"] == "Grouped computation of tmax and tmin"

    with pytest.raises(ValueError, match="Output #2 is missing a var_name!"):
        ind = Daily(
            realm="atmos",
            identifier="minmaxtemp2",
            cf_attrs=[
                dict(
                    var_name="tmin",
                    units="K",
                ),
                dict(
                    units="K",
                ),
            ],
            compute=multitemp_compute,
        )

    # Attrs passed as keywords - individually
    ind = Daily(
        realm="atmos",
        identifier="minmaxtemp3",
        var_name=["tmin", "tmax"],
        units="K",
        standard_name=["Min temp", ""],
        description="Grouped computation of tmax and tmin",
        compute=multitemp_compute,
    )
    tmin, tmax = ind(tas, freq="YS")
    assert tmin[0] == tas.min()
    assert tmax[0] == tas.max()
    assert tmin.attrs["standard_name"] == "Min temp"
    assert tmin.attrs["description"] == "Grouped computation of tmax and tmin"
    assert tmax.attrs["description"] == "Grouped computation of tmax and tmin"
    assert ind.units == ["K", "K"]

    # All must be the same length
    with pytest.raises(ValueError, match="Attribute var_name has 2 elements"):
        ind = Daily(
            realm="atmos",
            identifier="minmaxtemp3",
            var_name=["tmin", "tmax"],
            units="K",
            standard_name=["Min temp"],
            description="Grouped computation of tmax and tmin",
            compute=uniindpr_compute,
        )

    ind = Daily(
        realm="atmos",
        identifier="minmaxtemp4",
        var_name=["tmin", "tmax"],
        units="K",
        standard_name=["Min temp", ""],
        description="Grouped computation of tmax and tmin",
        compute=uniindtemp_compute,
    )
    with pytest.raises(ValueError, match="Indicator minmaxtemp4 was wrongly defined"):
        _tmin, _tmax = ind(tas, freq="YS")


def test_missing(tas_series):
    a = tas_series(np.ones(365, float), start="1/1/2000")

    # By default, missing is set to "from_context", and the default missing option is "any"
    # Cannot set missing_options with "from_context"
    with pytest.raises(ValueError, match="Cannot set `missing_options`"):
        uniClim.__class__(missing_options={"tolerance": 0.01})

    # Null value
    a[5] = np.nan

    m = uniIndTemp(a, freq="MS")
    assert m[0].isnull()

    with xclim.set_options(
        check_missing="pct", missing_options={"pct": {"tolerance": 0.05}}
    ):
        m = uniIndTemp(a, freq="MS")
        assert not m[0].isnull()
        assert "check_missing=pct, missing_options={'tolerance': 0.05}" in m.history

    with xclim.set_options(check_missing="wmo"):
        m = uniIndTemp(a, freq="YS")
        assert m[0].isnull()

    # With freq=None
    c = uniClim(a)
    assert c.isnull()

    # With indexer
    ci = uniClim(a, month=[2])
    assert not ci.isnull()

    out = uniClim(a, month=[1])
    assert out.isnull()


def test_missing_from_context(tas_series):
    a = tas_series(np.ones(365, float), start="1/1/2000")
    # Null value
    a[5] = np.nan

    ind = uniIndTemp.__class__(missing="from_context")

    m = ind(a, freq="MS")
    assert m[0].isnull()


def test_json(pr_series):
    meta = uniIndPr.json()

    expected = {
        "identifier",
        "title",
        "keywords",
        "abstract",
        "parameters",
        "history",
        "references",
        "notes",
        "outputs",
    }

    output_exp = {
        "var_name",
        "units",
        "long_name",
        "standard_name",
        "cell_methods",
        "description",
        "comment",
    }

    assert set(meta.keys()).issubset(expected)
    for output in meta["outputs"]:
        assert set(output.keys()).issubset(output_exp)


def test_all_jsonable(official_indicators):
    problems = []
    err = None
    for identifier, ind in official_indicators.items():
        indinst = ind.get_instance()
        json.dumps(indinst.json())
        try:
            json.dumps(indinst.json())
        except (KeyError, TypeError) as e:
            problems.append(identifier)
            err = e
    if problems:
        raise ValueError(
            f"Indicators {problems} provide problematic json serialization.: {err}"
        )


def test_all_parameters_understood(official_indicators):
    problems = set()
    for identifier, ind in official_indicators.items():
        indinst = ind.get_instance()
        for name, param in indinst.parameters.items():
            if param.kind == InputKind.OTHER_PARAMETER:
                problems.add((identifier, name))
    # this one we are ok with.
    if problems - {
        ("COOL_NIGHT_INDEX", "lat"),
        ("DRYNESS_INDEX", "lat"),
        # TODO: How should we handle the case of Literal[str]?
        ("GROWING_SEASON_END", "op"),
        ("GROWING_SEASON_START", "op"),
    }:
        raise ValueError(
            f"The following indicator/parameter couple {problems} use types not listed in InputKind."
        )


def test_signature():
    sig = signature(xclim.atmos.solid_precip_accumulation)
    assert list(sig.parameters.keys()) == [
        "pr",
        "tas",
        "thresh",
        "freq",
        "ds",
        "indexer",
    ]
    assert sig.parameters["pr"].annotation == xr.DataArray | str
    assert sig.parameters["tas"].default == "tas"
    assert sig.parameters["tas"].kind == sig.parameters["tas"].POSITIONAL_OR_KEYWORD
    assert sig.parameters["thresh"].kind == sig.parameters["thresh"].KEYWORD_ONLY
    assert sig.return_annotation == xr.DataArray

    sig = signature(xclim.atmos.wind_speed_from_vector)
    assert sig.return_annotation == tuple[xr.DataArray, xr.DataArray]


def test_doc():
    doc = xclim.atmos.cffwis_indices.__doc__
    assert doc.startswith("Canadian Fire Weather Index System indices. (realm: atmos)")
    assert "This indicator will check for missing values according to the method" in doc
    assert (
        "Based on indice :py:func:`~xclim.indices.fire._cffwis.cffwis_indices`." in doc
    )
    assert "ffmc0 : str or DataArray, optional" in doc
    assert "Returns\n-------" in doc
    assert "See :cite:t:`code-natural_resources_canada_data_nodate`, " in doc
    assert "the :py:mod:`xclim.indices.fire` module documentation," in doc
    assert (
        "and the docstring of :py:func:`fire_weather_ufunc` for more information."
        in doc
    )


def test_delayed(tasmax_series):
    tasmax = tasmax_series(np.arange(360.0)).chunk({"time": 5})
    out = uniIndTemp(tasmax)
    assert isinstance(out.data, dask.array.Array)


def test_identifier():
    with pytest.warns(UserWarning):
        uniIndPr.__class__(identifier="t_{}")


def test_formatting(pr_series):
    out = atmos.wetdays(pr_series(np.arange(366)), thresh=1.0 * units.mm / units.day)
    # pint 0.10 now pretty print day as d.
    assert (
        out.attrs["long_name"]
        == "Number of days with daily precipitation at or above 1 mm/d"
    )
    assert out.attrs["description"] in [
        "Annual number of days with daily precipitation at or above 1 mm/d."
    ]
    out = atmos.wetdays(pr_series(np.arange(366)), thresh=1.5 * units.mm / units.day)
    assert (
        out.attrs["long_name"]
        == "Number of days with daily precipitation at or above 1.5 mm/d"
    )
    assert out.attrs["description"] in [
        "Annual number of days with daily precipitation at or above 1.5 mm/d."
    ]


def test_parse_doc():
    doc = parse_doc(tg_mean.__doc__)
    assert doc["title"] == "Mean of daily average temperature."
    assert (
        doc["abstract"]
        == "Resample the original daily mean temperature series by taking the mean over each period."
    )
    assert doc["parameters"]["tas"]["description"] == "Mean daily temperature."
    assert doc["parameters"]["freq"]["description"] == "Resampling frequency."
    assert doc["notes"].startswith("Let")
    assert "math::" in doc["notes"]
    assert "references" not in doc
    assert doc["long_name"] == "The mean daily temperature at the given time frequency"

    doc = parse_doc(xclim.indices.saturation_vapor_pressure.__doc__)
    assert (
        doc["parameters"]["ice_thresh"]["description"]
        == "Threshold temperature under which to switch to equations in reference to ice instead of water. "
        "If None (default) everything is computed with reference to water."
    )
    assert "goff_low-pressure_1946" in doc["references"]


def test_parsed_doc():
    assert "tas" in xclim.atmos.liquid_precip_accumulation.parameters

    params = xclim.atmos.drought_code.parameters
    assert params["tas"].description == "Noon temperature."
    assert params["tas"].units == "[temperature]"
    assert params["tas"].kind is InputKind.VARIABLE
    assert params["tas"].default == "tas"
    assert params["snd"].default is None
    assert params["snd"].kind is InputKind.OPTIONAL_VARIABLE
    assert params["snd"].units == "[length]"
    assert params["season_method"].kind is InputKind.STRING
    assert params["season_method"].choices == {"GFWED", None, "WF93", "LA08"}

    params = xclim.atmos.standardized_precipitation_evapotranspiration_index.parameters
    assert params["fitkwargs"].kind is InputKind.DICT


def test_default_formatter():
    assert default_formatter.format("{freq}", freq="YS") == "annual"
    assert default_formatter.format("{freq:noun}", freq="MS") == "months"
    assert default_formatter.format("{month}", month="m3") == "march"


def test_AttrFormatter():
    fmt = AttrFormatter(
        mapping={"evil": ["méchant", "méchante"], "nice": ["beau", "belle"]},
        modifiers=["m", "f"],
    )
    # Normal cases
    assert fmt.format("{adj:m}", adj="evil") == "méchant"
    assert fmt.format("{adj:f}", adj="nice") == "belle"
    # Missing mod:
    assert fmt.format("{adj}", adj="evil") == "méchant"
    # Mod with unknown value
    with pytest.warns(match="Requested formatting `m` for unknown string `funny`."):
        fmt.format("{adj:m}", adj="funny")


@pytest.mark.parametrize("new_line", ["<>", "\n"])
@pytest.mark.parametrize("missing_str", ["<Missing>", None])
def test_merge_attributes(missing_str, new_line):
    a = xr.DataArray([0], attrs={"text": "Text1"}, name="a")
    b = xr.DataArray([0], attrs={})
    c = xr.Dataset(attrs={"text": "Text3"})

    merged = merge_attributes(
        "text", a, missing_str=missing_str, new_line=new_line, b=b, c=c
    )

    assert merged.startswith("a: Text1")

    if missing_str is not None:
        assert merged.count(new_line) == 2
        assert f"b: {missing_str}" in merged
    else:
        assert merged.count(new_line) == 1
        assert "b:" not in merged


def test_update_history():
    a = xr.DataArray([0], attrs={"history": "Text1"}, name="a")
    b = xr.DataArray([0], attrs={"history": "Text2"})
    c = xr.Dataset(attrs={"history": "Text3"})

    merged = update_history("text", a, new_name="d", b=b, c=c)

    assert "d: text" in merged.split("\n")[-1]
    assert merged.startswith("a: Text1")


def test_input_dataset(open_dataset):
    ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")

    # Use defaults
    _ = xclim.atmos.daily_temperature_range(freq="YS", ds=ds)

    # Use non-defaults (inverted on purpose)
    with xclim.set_options(cf_compliance="log"):
        _ = xclim.atmos.daily_temperature_range("tasmax", "tasmin", freq="YS", ds=ds)

    # Use a mix
    _ = xclim.atmos.daily_temperature_range(tasmax=ds.tasmax, freq="YS", ds=ds)

    # Inexistent variable:
    dsx = ds.drop_vars("tasmin")
    with pytest.raises(MissingVariableError):
        out = xclim.atmos.daily_temperature_range(freq="YS", ds=dsx)  # noqa

    # dataset not given
    with pytest.raises(ValueError):
        xclim.atmos.daily_temperature_range(tasmax="tmax")


def test_indicator_from_dict():
    d = dict(
        realm="atmos",
        cf_attrs=dict(
            var_name="tmean{threshold}",
            units="K",
            long_name="{freq} mean surface temperature",
            standard_name="{freq} mean temperature",
            cell_methods=[{"time": "mean within days"}],
        ),
        compute="thresholded_statistics",
        parameters=dict(
            threshold={"description": "A threshold temp"},
            op="<",
            reducer="mean",
        ),
        input={"data": "tas"},
    )

    ind = Daily.from_dict(d, identifier="tmean", module="test")

    assert ind.realm == "atmos"
    # Parameters metadata modification
    assert ind.parameters["threshold"].description == "A threshold temp"
    # Injection of parameters
    assert ind.injected_parameters["op"] == "<"
    assert ind.parameters["tas"].compute_name == "data"
    assert signature(ind).parameters["tas"].default == "tas"
    assert ind.parameters["tas"].units == "[temperature]"

    # Wrap a multi-output ind
    d = dict(base="wind_speed_from_vector")
    Indicator.from_dict(d, identifier="wsfv", module="test")


def test_indicator_errors():
    def func(data: xr.DataArray, thresh: str = "0 degC", freq: str = "YS"):  # noqa
        return data

    doc = [
        "The title",
        "",
        "    The abstract",
        "",
        "    Parameters",
        "    ----------",
        "    data : xr.DataArray",
        "      A variable.",
        "    thresh : str",
        "      A threshold",
        "    freq : str",
        "      The resampling frequency.",
        "",
        "    Returns",
        "    -------",
        "    xr.DataArray, [K]",
        "      An output",
    ]
    func.__doc__ = "\n".join(doc)

    d = dict(
        realm="atmos",
        cf_attrs=dict(
            var_name="tmean{threshold}",
            units="K",
            long_name="{freq} mean surface temperature",
            standard_name="{freq} mean temperature",
            cell_methods=[{"time": "mean within days"}],
        ),
        compute=func,
        input={"data": "tas"},
    )
    ind = Daily(identifier="indi", module="test", **d)

    with pytest.raises(AttributeError, match="`identifier` has not been set"):
        Daily(**d)

    d["identifier"] = "bad_indi"
    d["module"] = "test"

    bad_doc = doc[:12] + ["    extra: bool", "      Woupsi"] + doc[12:]
    func.__doc__ = "\n".join(bad_doc)
    with pytest.raises(ValueError, match="Malformed docstring"):
        Daily(**d)

    func.__doc__ = "\n".join(doc)
    d["parameters"] = {}
    d["parameters"]["thresh"] = "1 degK"
    d["parameters"]["extra"] = "woopsi again"
    with pytest.raises(ValueError, match="Parameter 'extra' was passed but it does"):
        Daily(**d)

    del d["parameters"]["extra"]
    d["input"]["data"] = "3nsd6sk72"
    with pytest.raises(ValueError, match="Compute argument data was mapped to"):
        Daily(**d)

    d2 = dict(input={"tas": "sfcWind"})
    with pytest.raises(ValueError, match="When changing the name of a variable by"):
        ind.__class__(**d2)

    del d["input"]
    # with pytest.raises(ValueError, match="variable data is missing expected units"):
    #     Daily(**d)

    d["parameters"]["thresh"] = {"units": "K"}
    d["realm"] = "mercury"
    d["input"] = {"data": "tasmin"}
    with pytest.raises(AttributeError, match="Indicator's realm must be given as one"):
        Daily(**d)

    def func(data: xr.DataArray, thresh: str = "0 degC"):
        return data

    func.__doc__ = "\n".join(doc[:10] + doc[12:])
    d = dict(
        realm="atmos",
        cf_attrs=dict(
            var_name="tmean{threshold}",
            units="K",
            long_name="{freq} mean surface temperature",
            standard_name="{freq} mean temperature",
            cell_methods=[{"time": "mean within days"}],
        ),
        compute=func,
        input={"data": "tas"},
    )
    with pytest.raises(ValueError, match="ResamplingIndicator require a 'freq'"):
        Daily(identifier="indi", module="test", **d)


def test_indicator_call_errors(tas_series):
    tas = tas_series(np.arange(730), start="2001-01-01")
    uniIndTemp(da=tas, thresh="3 K")

    with pytest.raises(TypeError, match="too many positional arguments"):
        uniIndTemp(tas, tas)

    with pytest.raises(TypeError, match="got an unexpected keyword argument 'oups'"):
        uniIndTemp(tas, oups=3)


def test_resamplingIndicator_new_error():
    with pytest.raises(ValueError, match="ResamplingIndicator require a 'freq'"):
        Daily(
            realm="atmos",
            identifier="multiopt",
            cf_attrs=[dict(units="K")],
            module="test",
            compute=multioptvar_compute,
        )


def test_resampling_indicator_with_indexing(tas_series):
    tas = tas_series(np.ones(731) + 273.15, start="2003-01-01")

    out = xclim.atmos.tx_days_above(tas, thresh="0 degC", freq="YS")
    np.testing.assert_allclose(out, [365, 366])

    out = xclim.atmos.tx_days_above(tas, thresh="0 degC", freq="YS", month=2)
    np.testing.assert_allclose(out, [28, 29])

    out = xclim.atmos.tx_days_above(
        tas, thresh="0 degC", freq="YS-JUL", doy_bounds=(1, 50)
    )
    np.testing.assert_allclose(out, [50, 50, np.nan])

    out = xclim.atmos.tx_days_above(
        tas, thresh="0 degC", freq="YS", date_bounds=("02-29", "04-01")
    )
    np.testing.assert_allclose(out, [32, 33])


def test_all_inputs_known():
    var_and_inds = list_input_variables()
    known_vars = (
        set(var_and_inds.keys())
        - {
            "dc0",
            "season_mask",
            "ffmc0",
            "dmc0",
            "kbdi0",
            "drought_factor",
        }  # FWI optional inputs
        - {var for var in var_and_inds.keys() if var.endswith("_per")}  # percentiles
        - {"pr_annual", "pr_cal", "wb_cal"}  # other optional or uncommon
        - {"q", "da"}  # Generic inputs
        - {"mrt", "wb"}  # TODO: add Mean Radiant Temperature and water budget
    )
    if not set(VARIABLES.keys()).issuperset(known_vars):
        raise AssertionError(
            "All input variables of xclim indicators must be registered in "
            "data/variables.yml, or skipped explicitly in this test. "
            f"The yaml file is missing: {known_vars - VARIABLES.keys()}."
        )


def test_freq_doc():
    from xclim import atmos

    doc = atmos.latitude_temperature_index.__doc__
    allowed_periods = ["Y"]
    exp = f"Restricted to frequencies equivalent to one of {allowed_periods}"
    assert exp in doc



from __future__ import annotations

import platform
from pathlib import Path

import numpy as np
import pytest
from xarray import Dataset

from xclim import __version__ as __xclim_version__
from xclim.testing.helpers import test_timeseries as timeseries
from xclim.testing.utils import open_dataset, publish_release_notes, show_versions


class TestFixtures:
    def test_timeseries_made_up_variable(self):
        ds = timeseries(
            np.zeros(31),
            "luminiferous_aether_flux",
            units="W K mol A-1 m-2 s-1",
            as_dataset=True,
        )

        assert isinstance(ds, Dataset)
        assert ds.luminiferous_aether_flux.attrs["units"] == "W K mol A-1 m-2 s-1"
        assert "standard_name" not in ds.luminiferous_aether_flux.attrs


class TestFileRequests:

    @staticmethod
    def file_md5_checksum(f_name):
        import hashlib

        hash_md5 = hashlib.md5()  # noqa: S324
        with open(f_name, "rb") as f:
            hash_md5.update(f.read())
        return hash_md5.hexdigest()

    @pytest.mark.requires_internet
    def test_open_testdata(
        self,
    ):
        from xclim.testing.utils import default_testdata_cache, default_testdata_version

        # Test with top-level default engine
        ds = open_dataset(
            Path("cmip5/tas_Amon_CanESM2_rcp85_r1i1p1_200701-200712.nc"),
            cache_dir=default_testdata_cache.joinpath(default_testdata_version),
            engine="h5netcdf",
        )
        assert ds.lon.size == 128

    def test_md5_sum(self):
        test_data = Path(__file__).parent / "data"
        callendar = test_data / "callendar_1938.txt"
        md5_sum = self.file_md5_checksum(callendar)
        if platform.system() == "Windows":
            # Windows has a different line ending (CR-LF) than Unix (LF)
            assert md5_sum == "38083271c2d4c85dea6bd6baf23d34de"  # noqa
        else:
            assert md5_sum == "9a5d9f94d76d4f9d9b7aaadbe8cbf541"  # noqa


class TestReleaseSupportFuncs:
    def test_show_version_file(self, tmp_path):
        temp_filename = tmp_path.joinpath("version_info.txt")
        show_versions(file=temp_filename)

        with open(temp_filename) as f:
            contents = f.readlines().copy()
            assert "INSTALLED VERSIONS\n" in contents
            assert "------------------\n" in contents
            assert f"python: {platform.python_version()}\n" in contents
            assert f"xclim: {__xclim_version__}\n" in contents
            assert "boltons: installed\n" in contents

    @pytest.mark.requires_docs
    def test_release_notes_file(self, tmp_path):
        temp_filename = tmp_path.joinpath("version_info.txt")
        publish_release_notes(style="md", file=temp_filename)

        with open(temp_filename) as f:
            assert "# Changelog" in f.readlines()[0]

    @pytest.mark.requires_docs
    def test_release_notes_file_not_implemented(self, tmp_path):
        temp_filename = tmp_path.joinpath("version_info.txt")
        with pytest.raises(NotImplementedError):
            publish_release_notes(style="qq", file=temp_filename)



"""
Indices Helper Functions Submodule
==================================

Functions that encapsulate logic and can be shared by many indices,
but are not particularly index-like themselves (those should go in the :py:mod:`xclim.indices.generic` module).
"""

from __future__ import annotations

from collections.abc import Callable, Mapping
from datetime import timedelta
from inspect import stack
from typing import Any, Literal, cast

import cf_xarray  # noqa: F401, pylint: disable=unused-import
import cftime
import numba as nb
import numpy as np
import xarray as xr

try:
    from xarray.coding.calendar_ops import (
        _datetime_to_decimal_year as datetime_to_decimal_year,
    )
except ImportError:
    XR2409 = True
else:
    XR2409 = False

from xclim.core import Quantified
from xclim.core.calendar import ensure_cftime_array, get_calendar
from xclim.core.options import MAP_BLOCKS, OPTIONS
from xclim.core.units import convert_units_to
from xclim.core.utils import _chunk_like, uses_dask


def _wrap_radians(da):
    with xr.set_options(keep_attrs=True):
        return ((da + np.pi) % (2 * np.pi)) - np.pi


def distance_from_sun(dates: xr.DataArray) -> xr.DataArray:
    """
    Sun-earth distance.

    The distance from sun to earth in astronomical units.

    Parameters
    ----------
    dates : xr.DataArray
        Series of dates and time of days.

    Returns
    -------
    xr.DataArray, [astronomical units]
        Sun-earth distance.

    References
    ----------
    # TODO: Find a way to reference this
    U.S. Naval Observatory:Astronomical Almanac. Washington, D.C.: U.S. Government Printing Office (1985).
    """
    cal = get_calendar(dates)
    if cal == "default":
        cal = "standard"
    days_since = cftime.date2num(
        ensure_cftime_array(dates), "days since 2000-01-01 12:00:00", calendar=cal
    )
    g = ((357.528 + 0.9856003 * days_since) % 360) * np.pi / 180
    sun_earth = 1.00014 - 0.01671 * np.cos(g) - 0.00014 * np.cos(2.0 * g)
    return xr.DataArray(sun_earth, coords=dates.coords, dims=dates.dims)


def day_angle(time: xr.DataArray):
    """Day of year as an angle.

    Assuming the earth makes a full circle in a year, this is the angle covered from
    the beginning of the year up to that timestep. Also called the "julian day fraction".
    See :py:func:`~xclim.core.calendar.datetime_to_decimal_year`.
    """
    if XR2409:
        decimal_year = time.dt.decimal_year
    else:
        decimal_year = datetime_to_decimal_year(times=time, calendar=time.dt.calendar)
    return ((decimal_year % 1) * 2 * np.pi).assign_attrs(units="rad")


def solar_declination(time: xr.DataArray, method="spencer") -> xr.DataArray:
    """Solar declination.

    The angle between the sun rays and the earth's equator, in radians, as approximated
    by :cite:t:`spencer_fourier_1971` or assuming the orbit is a circle.

    Parameters
    ----------
    time: xr.DataArray
      Time coordinate.
    method : {'spencer', 'simple'}
      Which approximation to use. The default ("spencer") uses the first 7 terms of the
      Fourier series representing the observed declination, while "simple" assumes
      the orbit is a circle with a fixed obliquity and that the solstice/equinox happen
      at fixed angles on the orbit (the exact calendar date changes for leap years).

    Returns
    -------
    xr.DataArray, [rad]
        Solar declination angle.

    References
    ----------
    :cite:cts:`spencer_fourier_1971`
    """
    # julian day fraction
    da = convert_units_to(day_angle(time), "rad")
    if method == "simple":
        # This assumes the orbit is a perfect circle, the obliquity is 0.4091 rad (23.43°)
        # and the equinox is on the March 21st 17:20 UTC (March 20th 23:14 UTC on leap years)
        sd = 0.4091 * np.sin(da - 1.39)
    elif method == "spencer":
        sd = (
            0.006918
            - 0.399912 * np.cos(da)
            + 0.070257 * np.sin(da)
            - 0.006758 * np.cos(2 * da)
            + 0.000907 * np.sin(2 * da)
            - 0.002697 * np.cos(3 * da)
            + 0.001480 * np.sin(3 * da)
        )
    else:
        raise NotImplementedError(
            f"Method {method} must be one of 'simple' or 'spencer'"
        )
    return _wrap_radians(sd).assign_attrs(units="rad").rename("declination")


def time_correction_for_solar_angle(time: xr.DataArray) -> xr.DataArray:
    """Time correction for solar angle.

    Every 1° of angular rotation on earth is equal to 4 minutes of time.
    The time correction is needed to adjust local watch time to solar time.

    Parameters
    ----------
    time: xr.DataArray
      Time coordinate.

    Returns
    -------
    xr.DataArray, [rad]
        Time correction of solar angle.

    References
    ----------
    :cite:cts:`di_napoli_mean_2020`
    """
    da = convert_units_to(day_angle(time), "rad")
    tc = (
        0.004297
        + 0.107029 * np.cos(da)
        - 1.837877 * np.sin(da)
        - 0.837378 * np.cos(2 * da)
        - 2.340475 * np.sin(2 * da)
    )
    tc = tc.assign_attrs(units="degrees")
    return _wrap_radians(convert_units_to(tc, "rad"))


def eccentricity_correction_factor(
    time: xr.DataArray, method: str = "spencer"
) -> xr.DataArray:
    """Eccentricity correction factor of the Earth's orbit.

    The squared ratio of the mean distance Earth-Sun to the distance at a specific moment.
    As approximated by :cite:t:`spencer_fourier_1971`.

    Parameters
    ----------
    time: xr.DataArray
        Time coordinate
    method : {'spencer', 'simple'}
        Which approximation to use.
        The default ("spencer") uses the first five terms of the fourier series of the eccentricity.
        The "simple" method approximates with only the first two.

    Returns
    -------
    xr.DataArray, [dimensionless]
        Eccentricity correction factor.

    References
    ----------
    :cite:cts:`spencer_fourier_1971,perrin_estimation_1975`
    """
    # julian day fraction
    da = convert_units_to(day_angle(time), "rad")
    if method == "simple":
        # It is quite used, I think the source is (not available online):
        # Perrin de Brichambaut, C. (1975).
        # Estimation des ressources énergétiques solaires en France. Ed. Européennes thermique et industrie.
        return cast(xr.DataArray, 1 + 0.033 * np.cos(da))
    if method == "spencer":
        return cast(
            xr.DataArray,
            1.0001100
            + 0.034221 * np.cos(da)
            + 0.001280 * np.sin(da)
            + 0.000719 * np.cos(2 * da)
            + 0.000077 * np.sin(2 * da),
        )
    raise NotImplementedError("Method must be one of 'simple' or 'spencer'.")


def cosine_of_solar_zenith_angle(
    time: xr.DataArray,
    declination: xr.DataArray,
    lat: Quantified,
    lon: Quantified = "0 °",
    time_correction: xr.DataArray | None = None,
    stat: str = "average",
    sunlit: bool = False,
    chunks: dict[str, int] | None = None,
) -> xr.DataArray:
    """Cosine of the solar zenith angle.

    The solar zenith angle is the angle between a vertical line (perpendicular to the ground) and the sun rays.
    This function computes a statistic of its cosine : its instantaneous value, the integral from sunrise to sunset or the average over
    the same period or over a subdaily interval.
    Based on :cite:t:`kalogirou_chapter_2014` and :cite:t:`di_napoli_mean_2020`.

    Parameters
    ----------
    time: xr.DataArray
        The UTC time. If not daily and `stat` is "integral" or "average", the timestamp is taken as the start of interval.
        If daily, the interval is assumed to be centered on Noon.
        If fewer than three timesteps are given, a daily frequency is assumed.
    declination : xr.DataArray
        Solar declination. See :py:func:`solar_declination`.
    lat : Quantified
        Latitude.
    lon : Quantified
        Longitude. Needed if the input timeseries is subdaily.
    time_correction : xr.DataArray, optional
        Time correction for solar angle. See :py:func:`time_correction_for_solar_angle`
        This is necessary if stat is "instant".
    stat : {'average', 'integral', 'instant'}
        Which daily statistic to return.
        If "average", this returns the average of the cosine of the zenith angle
        If "integral", this returns the integral of the cosine of the zenith angle
        If "instant", this returns the instantaneous cosine of the zenith angle
    sunlit: bool
        If True, only the sunlit part of the interval is considered in the integral or average.
        Does nothing if stat is "instant".
    chunks : dictionary
        When `time`,  `lat` and `lon` originate from coordinates of a large chunked dataset, this dataset's chunking
        can be passed here to ensure the computation is also chunked.

    Returns
    -------
    xr.DataArray, [rad] or [dimensionless]
        Cosine of the solar zenith angle. If stat is "integral", dimensions can be said to be "time" as the integral
        is on the hour angle.
        For seconds, multiply by the number of seconds in a complete day cycle (24*60*60) and divide by 2π.

    Notes
    -----
    This code was inspired by the `thermofeel` and `PyWBGT` package.

    References
    ----------
    :cite:cts:`kalogirou_chapter_2014,di_napoli_mean_2020`
    """
    declination = convert_units_to(declination, "rad")
    lat = _wrap_radians(convert_units_to(lat, "rad"))
    lon = convert_units_to(lon, "rad")
    declination, lat, lon = _chunk_like(declination, lat, lon, chunks=chunks)

    S_IN_D = 24 * 3600

    if len(time) < 3 or xr.infer_freq(time) == "D":
        h_s = -np.pi if stat != "instant" else 0
        h_e = np.pi - 1e-9  # just below pi
    else:
        if time.dtype == "O":  # cftime
            time_as_s = time.copy(
                data=xr.CFTimeIndex(cast(time.values, np.ndarray)).asi8 / 1e6
            )
        else:  # numpy
            time_as_s = time.copy(data=time.astype(float) / 1e9)
        h_s_utc = (((time_as_s % S_IN_D) / S_IN_D) * 2 * np.pi + np.pi).assign_attrs(
            units="rad"
        )
        h_s = h_s_utc + lon

        interval_as_s = time.diff("time").dt.seconds.reindex(
            time=time.time, method="bfill"
        )
        h_e = h_s + 2 * np.pi * interval_as_s / S_IN_D

    if stat == "instant":
        h_s = h_s + time_correction

        return cast(
            xr.DataArray,
            np.sin(declination) * np.sin(lat)
            + np.cos(declination) * np.cos(lat) * np.cos(h_s),
        ).clip(0, None)
    if stat not in {"average", "integral"}:
        raise NotImplementedError(
            "Argument 'stat' must be one of 'integral', 'average' or 'instant'."
        )
    if sunlit:
        # hour angle of sunset (eq. 2.15), with NaNs inside the polar day/night
        tantan = cast(xr.DataArray, -np.tan(lat) * np.tan(declination))
        h_ss = np.arccos(tantan.where(abs(tantan) <= 1))
    else:
        # Whole period, so we put sunset at midnight
        h_ss = np.pi - 1e-9

    return xr.apply_ufunc(
        _sunlit_integral_of_cosine_of_solar_zenith_angle,
        declination,
        lat,
        _wrap_radians(h_ss),
        _wrap_radians(h_s),
        _wrap_radians(h_e),
        stat == "average",
        input_core_dims=[[]] * 6,
        dask="parallelized",
    )


@nb.vectorize
def _sunlit_integral_of_cosine_of_solar_zenith_angle(
    declination, lat, h_sunset, h_start, h_end, average
):
    """Integral of the cosine of the the solar zenith angle over the sunlit part of the interval."""
    # Code inspired by PyWBGT
    h_sunrise = -h_sunset
    # Polar day
    if np.isnan(h_sunset) & ((declination * lat) > 0):
        num = np.sin(h_end) - np.sin(h_start)
        # Polar day with interval crossing midnight
        if h_end < h_start:
            denum = h_end + 2 * np.pi - h_start
        else:
            denum = h_end - h_start
    # Polar night:
    elif np.isnan(h_sunset) & ((declination * lat) < 0):
        return 0
    # No sunlit interval (at night) 1) crossing midnight and 2) between 0h and sunrise 3) between sunset and 0h
    elif (
        (h_start > h_sunset and h_end < h_sunrise)
        or (h_start < h_sunrise and h_end < h_sunrise)
        or (h_start > h_sunset and h_end > h_sunset)
    ):
        return 0
    # Interval crossing midnight, starting after sunset (before midnight), finishing after sunrise
    elif h_start > h_end >= h_sunrise and h_start >= h_sunset:
        num = np.sin(h_end) - np.sin(h_sunrise)
        denum = h_end - h_sunrise
    # Interval crossing midnight, starting after sunrise, finishing after sunset (after midnight)
    elif h_end < h_start and h_start >= h_sunrise >= h_end:
        num = np.sin(h_sunset) - np.sin(h_start)
        denum = h_sunset - h_start
    # Interval crossing midnight, starting before sunset, finishing after sunrise (2 sunlit parts)
    elif h_sunset >= h_start > h_end >= h_sunrise:
        num = np.sin(h_sunset) - np.sin(h_start) + np.sin(h_end) - np.sin(h_sunrise)
        denum = h_sunset - h_start + h_end - h_sunrise
    # All other cases : interval not crossing midnight, overlapping with the sunlit part
    else:
        h1 = max(h_sunrise, h_start)
        h2 = min(h_sunset, h_end)
        num = np.sin(h2) - np.sin(h1)
        denum = h2 - h1
    out = (
        np.sin(declination) * np.sin(lat) * denum
        + np.cos(declination) * np.cos(lat) * num
    )
    if average:
        out = out / denum
    return out


def extraterrestrial_solar_radiation(
    times: xr.DataArray,
    lat: xr.DataArray,
    solar_constant: Quantified = "1361 W m-2",
    method: str = "spencer",
    chunks: Mapping[Any, tuple] | None = None,
) -> xr.DataArray:
    """Extraterrestrial solar radiation.

    This is the daily energy received on a surface parallel to the ground at the mean distance of the earth to the sun.
    It neglects the effect of the atmosphere. Computation is based on :cite:t:`kalogirou_chapter_2014` and the default
    solar constant is taken from :cite:t:`matthes_solar_2017`.

    Parameters
    ----------
    times : xr.DataArray
        Daily datetime data. This function makes no sense with data of other frequency.
    lat : xr.DataArray
        Latitude.
    solar_constant : str
        The solar constant, the energy received on earth from the sun per surface per time.
    method : {'spencer', 'simple'}
        Which method to use when computing the solar declination and the eccentricity
        correction factor. See :py:func:`solar_declination` and :py:func:`eccentricity_correction_factor`.
    chunks : dict
        When `times` and `lat` originate from coordinates of a large chunked dataset, passing the dataset's chunks here
        will ensure the computation is chunked as well.

    Returns
    -------
    Extraterrestrial solar radiation, [J m-2 d-1]

    References
    ----------
    :cite:cts:`kalogirou_chapter_2014,matthes_solar_2017`
    """
    dr = eccentricity_correction_factor(times, method=method)
    ds = solar_declination(times, method=method)
    gsc = convert_units_to(solar_constant, "J m-2 d-1")
    rad_to_day = 1 / (2 * np.pi)  # convert radians of the "day circle" to day
    return (
        gsc
        * rad_to_day
        * cosine_of_solar_zenith_angle(
            times, ds, lat, stat="integral", sunlit=True, chunks=chunks
        )
        * dr
    ).assign_attrs(units="J m-2 d-1")


def day_lengths(
    dates: xr.DataArray,
    lat: xr.DataArray,
    method: str = "spencer",
) -> xr.DataArray:
    r"""Day-lengths according to latitude and day of year.

    See :py:func:`solar_declination` for the approximation used to compute the solar declination angle.
    Based on :cite:t:`kalogirou_chapter_2014`.

    Parameters
    ----------
    dates: xr.DataArray
        Daily datetime data. This function makes no sense with data of other frequency.
    lat: xarray.DataArray
        Latitude coordinate.
    method : {'spencer', 'simple'}
        Which approximation to use when computing the solar declination angle.
        See :py:func:`solar_declination`.

    Returns
    -------
    xarray.DataArray, [hours]
        Day-lengths in hours per individual day.

    References
    ----------
    :cite:cts:`kalogirou_chapter_2014`
    """
    declination = solar_declination(dates.time, method=method)
    lat = convert_units_to(lat, "rad")
    # arccos gives the hour-angle at sunset, multiply by 24 / 2π to get hours.
    # The day length is twice that.
    with np.errstate(invalid="ignore"):
        day_length_hours = (
            (24 / np.pi) * np.arccos(-np.tan(lat) * np.tan(declination))
        ).assign_attrs(units="h")

    return day_length_hours


def wind_speed_height_conversion(
    ua: xr.DataArray,
    h_source: str,
    h_target: str,
    method: str = "log",
) -> xr.DataArray:
    r"""Wind speed at two meters.

    Parameters
    ----------
    ua : xarray.DataArray
        Wind speed at height h
    h_source : str
        Height of the input wind speed `ua` (e.g. `h == "10 m"` for a wind speed at `10 meters`)
    h_target : str
        Height of the output wind speed
    method : {"log"}
        Method used to convert wind speed from one height to another

    Returns
    -------
    xarray.DataArray
        Wind speed at height `h_target`

    References
    ----------
    :cite:cts:`allen_crop_1998`
    """
    h_source = convert_units_to(h_source, "m")
    h_target = convert_units_to(h_target, "m")
    if method == "log":
        if min(h_source, h_target) < 1 + 5.42 / 67.8:
            raise ValueError(
                f"The height {min(h_source, h_target)}m is too small for method {method}. Heights must be greater than {1 + 5.42 / 67.8}"
            )
        with xr.set_options(keep_attrs=True):
            return ua * np.log(67.8 * h_target - 5.42) / np.log(67.8 * h_source - 5.42)
    else:
        raise NotImplementedError(f"'{method}' method is not implemented.")


def _gather_lat(da: xr.DataArray) -> xr.DataArray:
    """Gather latitude coordinate using cf-xarray.

    Parameters
    ----------
    da : xarray.DataArray
        CF-conformant DataArray with a "latitude" coordinate.

    Returns
    -------
    xarray.DataArray
        Latitude coordinate.
    """
    try:
        lat = da.cf["latitude"]
        return lat
    except KeyError as err:
        n_func = stack()[1].function
        msg = (
            f"{n_func} could not find latitude coordinate in DataArray. "
            "Try passing it explicitly (`lat=ds.lat`)."
        )
        raise ValueError(msg) from err


def _gather_lon(da: xr.DataArray) -> xr.DataArray:
    """Gather longitude coordinate using cf-xarray.

    Parameters
    ----------
    da : xarray.DataArray
        CF-conformant DataArray with a "longitude" coordinate.

    Returns
    -------
    xarray.DataArray
        Longitude coordinate.
    """
    try:
        lat = da.cf["longitude"]
        return lat
    except KeyError as err:
        n_func = stack()[1].function
        msg = (
            f"{n_func} could not find longitude coordinate in DataArray. "
            "Try passing it explicitly (`lon=ds.lon`)."
        )
        raise ValueError(msg) from err


def resample_map(
    obj: xr.DataArray | xr.Dataset,
    dim: str,
    freq: str,
    func: Callable,
    map_blocks: bool | Literal["from_context"] = "from_context",
    resample_kwargs: dict | None = None,
    map_kwargs: dict | None = None,
) -> xr.DataArray | xr.Dataset:
    r"""Wraps xarray's resample(...).map() with a :py:func:`xarray.map_blocks`, ensuring the chunking is appropriate using flox.

    Parameters
    ----------
    obj : DataArray or Dataset
        The xarray object to resample.
    dim : str
        Dimension over which to resample.
    freq : str
        Resampling frequency along `dim`.
    func : callable
        Function to map on each resampled group.
    map_blocks : bool or "from_context"
        If True, the resample().map() call is wrapped inside a `map_blocks`.
        If False, this does not do anything special.
        If "from_context", xclim's "resample_map_blocks" option is used.
        If the object is not using dask, this is set to False.
    resample_kwargs : dict, optional
        Other arguments to pass to `obj.resample()`.
    map_kwargs : dict, optional
        Arguments to pass to `map`.

    Returns
    -------
    Resampled object.
    """
    resample_kwargs = resample_kwargs or {}
    map_kwargs = map_kwargs or {}
    if map_blocks == "from_context":
        map_blocks = OPTIONS[MAP_BLOCKS]

    if not uses_dask(obj) or not map_blocks:
        return obj.resample({dim: freq}, **resample_kwargs).map(func, **map_kwargs)

    try:
        from flox.xarray import rechunk_for_blockwise
    except ImportError as err:
        msg = f"Using {MAP_BLOCKS}=True requires flox."
        raise ValueError(msg) from err

    # Make labels, a unique integer for each resample group
    labels = xr.full_like(obj[dim], -1, dtype=np.int32)
    for lbl, group_slice in enumerate(obj[dim].resample({dim: freq}).groups.values()):
        labels[group_slice] = lbl

    obj_rechunked = rechunk_for_blockwise(obj, dim, labels)

    def _resample_map(obj_chnk, dm, frq, rs_kws, fun, mp_kws):
        return obj_chnk.resample({dm: frq}, **rs_kws).map(fun, **mp_kws)

    # Template. We are hoping that this takes a negligeable time as it is never loaded.
    template = obj_rechunked.resample(**{dim: freq}, **resample_kwargs).first()

    # New chunks along time : infer the number of elements resulting from the resampling of each chunk
    if isinstance(obj_rechunked, xr.Dataset):
        chunksizes = obj_rechunked.chunks[dim]
    else:
        chunksizes = obj_rechunked.chunks[obj_rechunked.get_axis_num(dim)]
    new_chunks = []
    i = 0
    for chunksize in chunksizes:
        new_chunks.append(len(np.unique(labels[i : i + chunksize])))
        i += chunksize
    template = template.chunk({dim: tuple(new_chunks)})

    return obj_rechunked.map_blocks(
        _resample_map, (dim, freq, resample_kwargs, func, map_kwargs), template=template
    )


def _compute_daytime_temperature(
    hour_after_sunrise: xr.DataArray,
    tasmin: xr.DataArray,
    tasmax: xr.DataArray,
    daylength: xr.DataArray,
) -> xr.DataArray:
    """Compute daytime temperature based on a sinusoidal profile.

    Minimum temperature is reached at sunrise and maximum temperature 2h before sunset.

    Parameters
    ----------
    hours_after_sunrise : xarray.DataArray
        Hours after the last sunrise.
    tasmin : xarray.DataArray
        Daily minimum temperature.
    tasmax : xarray.DataArray
        Daily maximum temperature.
    daylength : xarray.DataArray
        Length of the day in hours.

    Returns
    -------
    xarray.DataArray
        Hourly daytime temperature.
    """
    return (tasmax - tasmin) * np.sin(
        (np.pi * hour_after_sunrise) / (daylength + 4)
    ) + tasmin


def _compute_nighttime_temperature(
    hours_after_sunset: xr.DataArray,
    tasmin: xr.DataArray,
    tas_sunset: xr.DataArray,
    daylength: xr.DataArray,
) -> xr.DataArray:
    """Compute nighttime temperature based on a logarithmic profile.

    Temperature at sunset is computed from previous daytime temperature,
    minimum temperature is reached at sunrise.

    Parameters
    ----------
    hours_after_sunset : xarray.DataArray
        Hours after the last sunset.
    tasmin : xarray.DataArray
        Daily minimum temperature.
    tas_sunset : xarray.DataArray
        Temperature at last sunset.
    daylength : xarray.DataArray
        Length of the day in hours.

    Returns
    -------
    xarray.DataArray
        Hourly nighttime temperature.
    """
    return tas_sunset - ((tas_sunset - tasmin) / np.log(24 - daylength)) * np.log(
        hours_after_sunset
    )


def _add_one_day(time: xr.DataArray) -> xr.DataArray:
    """Add one day to a time coordinate.

    Depending on the calendar/dtype of the time array we need to use numpy's or
    datetime's (for cftimes) timedelta.

    Parameters
    ----------
    time : xr.DataArray
        Time coordinate.

    Returns
    -------
    xr.DataArray
        Next day.
    """
    if time.dtype == "O":
        return time + timedelta(days=1)
    return time + np.timedelta64(1, "D")


def make_hourly_temperature(tasmin: xr.DataArray, tasmax: xr.DataArray) -> xr.DataArray:
    """Compute hourly temperatures from tasmin and tasmax.

    Based on the Linvill et al. "Calculating Chilling Hours and Chill Units from Daily
    Maximum and Minimum Temperature Observations", HortScience, 1990
    we assume a sinusoidal temperature profile during daylight and a logarithmic decrease after sunset
    with tasmin reached at sunsrise and tasmax reached 2h before sunset.

    For simplicity and because it's used for daily aggregation, we assume that sunrise globally happens at midnight
    and the sunsets after `daylength` hours computed via the :py:func:`day_lengths` function.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Daily minimum temperature.
    tasmax : xarray.DataArray
        Daily maximum temperature.

    Returns
    -------
    xarray.DataArray
        Hourly temperature.
    """
    data = xr.merge([tasmin, tasmax])
    data = data.assign_coords(time=data.time.dt.floor("D"))
    # We add one more timestamp so the resample function includes the last day
    data = xr.concat(
        [
            data,
            data.isel(time=-1).assign_coords(
                time=_add_one_day(data.isel(time=-1).time)
            ),
        ],
        dim="time",
    )

    daylength = day_lengths(data.time, data.lat)
    # Create daily chunks to avoid memory issues after the resampling
    data = data.assign(
        daylength=daylength,
        sunset_temp=_compute_daytime_temperature(
            daylength, data.tasmin, data.tasmax, daylength
        ),
        next_tasmin=data.tasmin.shift(time=-1),
    )
    # Compute hourly data by resampling and remove the last time stamp that was added earlier
    hourly = data.resample(time="h").ffill().isel(time=slice(0, -1))

    # To avoid "invalid value encountered in log" warning we set hours before sunset to 1
    nighttime_hours = nighttime_hours = (
        hourly.time.dt.hour + 1 - hourly.daylength
    ).clip(1)

    return xr.where(
        hourly.time.dt.hour < hourly.daylength,
        _compute_daytime_temperature(
            hourly.time.dt.hour, hourly.tasmin, hourly.tasmax, hourly.daylength
        ),
        _compute_nighttime_temperature(
            nighttime_hours,
            hourly.next_tasmin,
            hourly.sunset_temp,
            hourly.daylength - 1,
        ),
    ).assign_attrs(units=tasmin.units)



from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

from xclim.core import dataflags as df

K2C = 273.15


class TestDataFlags:
    @pytest.mark.parametrize(
        "vars_dropped, flags",
        [
            (["tasmin"], dict(tas_exceeds_tasmax=False, tas_below_tasmin=None)),
            (["tasmax"], dict(tas_exceeds_tasmax=None, tas_below_tasmin=False)),
            ([], dict(tas_exceeds_tasmax=False, tas_below_tasmin=False)),
        ],
    )
    def test_tas_temperature_flags(
        self, vars_dropped, flags, tas_series, tasmax_series, tasmin_series
    ):
        ds = xr.Dataset()
        for series, val in zip(
            [tas_series, tasmax_series, tasmin_series], [0, 10, -10], strict=False
        ):
            vals = val + K2C + np.sin(2 * np.pi * np.arange(366 * 3) / 366)
            arr = series(vals, start="1971-01-01")
            ds = xr.merge([ds, arr])

        ds = ds.drop_vars(vars_dropped)
        flagged_ds = df.data_flags(ds.tas, ds)

        np.testing.assert_equal(flagged_ds.temperature_extremely_high.values, False)
        np.testing.assert_equal(flagged_ds.temperature_extremely_low.values, False)
        np.testing.assert_equal(
            flagged_ds.values_repeating_for_5_or_more_days.values, False
        )
        np.testing.assert_equal(
            flagged_ds.outside_5_standard_deviations_of_climatology.values, False
        )

        for flag, val in flags.items():
            np.testing.assert_equal(getattr(flagged_ds, flag).values, val)

    def test_pr_precipitation_flags(self, pr_series):
        pr = pr_series(np.zeros(365), start="1971-01-01")
        pr += 1 / 3600 / 24
        pr[0:7] += 10 / 3600 / 24
        pr[-7:] += 11 / 3600 / 24

        flagged = df.data_flags(pr)
        np.testing.assert_equal(flagged.negative_accumulation_values.values, False)
        np.testing.assert_equal(flagged.very_large_precipitation_events.values, False)
        np.testing.assert_equal(
            flagged.values_eq_5_repeating_for_5_or_more_days.values,
            False,
        )
        np.testing.assert_equal(
            flagged.values_eq_1_repeating_for_10_or_more_days.values,
            False,
        )

    def test_suspicious_pr_data(self, pr_series):
        bad_pr = pr_series(np.zeros(365), start="1971-01-01")

        # Add some strangeness
        bad_pr[8] = -1e-6  # negative values
        bad_pr[120] = 301 / 3600 / 24  # 301mm/day
        bad_pr[121:141] = 1.1574074074074072e-05  # 1mm/day
        bad_pr[200:300] = 5.787037037037036e-05  # 5mm/day

        flagged = df.data_flags(bad_pr)
        np.testing.assert_equal(flagged.negative_accumulation_values.values, True)
        np.testing.assert_equal(flagged.very_large_precipitation_events.values, True)
        np.testing.assert_equal(
            flagged.values_eq_1_repeating_for_10_or_more_days.values, True
        )
        np.testing.assert_equal(
            flagged.values_eq_5_repeating_for_5_or_more_days.values, True
        )

    def test_suspicious_tas_data(self, tas_series, tasmax_series, tasmin_series):
        bad_ds = xr.Dataset()
        for series, val in zip(
            [tas_series, tasmax_series, tasmin_series], [0, 10, -10], strict=False
        ):
            vals = val + K2C + np.sin(2 * np.pi * np.arange(366 * 7) / 366)
            arr = series(vals, start="1971-01-01")
            bad_ds = xr.merge([bad_ds, arr])

        # Swap entire variable arrays
        bad_ds["tasmin"].values, bad_ds["tasmax"].values = (
            bad_ds.tasmax.values,
            bad_ds.tasmin.values,
        )

        bad_tas = bad_ds.tas.values
        # Add some jankiness to tas
        bad_tas[5] = 58 + K2C  # Fluke event beyond 5 standard deviations
        bad_tas[600:610] = 80 + K2C  # Repeating values above hot extreme
        bad_tas[950] = -95 + K2C  # Cold extreme
        bad_ds["tas"].values = bad_tas

        flagged = df.data_flags(bad_ds.tas, bad_ds)
        np.testing.assert_equal(flagged.temperature_extremely_high.values, True)
        np.testing.assert_equal(flagged.temperature_extremely_low.values, True)
        np.testing.assert_equal(
            flagged.values_repeating_for_5_or_more_days.values, True
        )
        np.testing.assert_equal(
            flagged.outside_5_standard_deviations_of_climatology.values,
            True,
        )
        np.testing.assert_equal(flagged.tas_exceeds_tasmax.values, True)
        np.testing.assert_equal(flagged.tas_below_tasmin.values, True)

    def test_raises(self, tasmax_series, tasmin_series):
        bad_ds = xr.Dataset()
        for series, val in zip([tasmax_series, tasmin_series], [10, -10], strict=False):
            vals = val + K2C + np.sin(2 * np.pi * np.arange(366 * 3) / 366)
            arr = series(vals, start="1971-01-01")
            bad_ds = xr.merge([bad_ds, arr])

        # At this stage, it should not raise an exception
        df.data_flags(bad_ds.tasmax, bad_ds, raise_flags=True)

        # Swap entire variable arrays
        bad_ds["tasmin"].values, bad_ds["tasmax"].values = (
            bad_ds.tasmax.values,
            bad_ds.tasmin.values,
        )
        with pytest.raises(
            df.DataQualityException,
            match="Maximum temperature values found below minimum temperatures.",
        ):
            df.data_flags(bad_ds.tasmax, bad_ds, raise_flags=True)

    def test_era5_ecad_qc_flag(self, open_dataset):
        bad_ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")  # noqa

        # Add some suspicious run values
        bad_ds["tas"].values[0][100:300] = 17 + K2C

        with pytest.raises(
            df.DataQualityException,
            match="Runs of repetitive values for 5 or more days found for tas.",
        ):
            df.ecad_compliant(bad_ds, raise_flags=True)

        df_flagged = df.ecad_compliant(bad_ds)
        np.testing.assert_array_equal(df_flagged.ecad_qc_flag, False)

    def test_names(self, pr_series):
        pr = pr_series(np.zeros(365), start="1971-01-01")
        flgs = df.data_flags(
            pr,
            flags={
                "values_op_thresh_repeating_for_n_or_more_days": {
                    "op": "==",
                    "n": 5,
                    "thresh": "-5.1 mm d-1",
                }
            },
        )
        assert (
            list(flgs.data_vars.keys())[0]
            == "values_eq_minus5point1_repeating_for_5_or_more_days"
        )



# noqa: D100
# This file is the setup for the doctest suite.
# This must be run using the following commands:
# python -c "from xclim.testing.utils import run_doctests; run_doctests()"

from __future__ import annotations

import os
from pathlib import Path

import pytest

from xclim.testing.helpers import (
    add_doctest_filepaths,
    add_example_file_paths,
    generate_atmos,
)
from xclim.testing.utils import (
    TESTDATA_BRANCH,
    TESTDATA_CACHE_DIR,
    TESTDATA_REPO_URL,
    gather_testing_data,
    testing_setup_warnings,
)
from xclim.testing.utils import nimbus as _nimbus
from xclim.testing.utils import open_dataset as _open_dataset


@pytest.fixture(autouse=True, scope="session")
def threadsafe_data_dir(tmp_path_factory):
    """Return a threadsafe temporary directory for storing testing data."""
    yield Path(tmp_path_factory.getbasetemp().joinpath("data"))


@pytest.fixture(scope="session")
def nimbus(threadsafe_data_dir, worker_id):
    """Return a nimbus object for the test data."""
    return _nimbus(
        repo=TESTDATA_REPO_URL,
        branch=TESTDATA_BRANCH,
        cache_dir=(
            TESTDATA_CACHE_DIR if worker_id == "master" else threadsafe_data_dir
        ),
    )


@pytest.fixture(scope="session")
def open_dataset(nimbus):
    """Return a function that opens a dataset from the test data."""

    def _open_session_scoped_file(file: str | os.PathLike, **xr_kwargs):
        xr_kwargs.setdefault("cache", True)
        xr_kwargs.setdefault("engine", "h5netcdf")
        return _open_dataset(
            file,
            branch=TESTDATA_BRANCH,
            repo=TESTDATA_REPO_URL,
            cache_dir=nimbus.path,
            **xr_kwargs,
        )

    return _open_session_scoped_file


@pytest.fixture(scope="session", autouse=True)
def is_matplotlib_installed(xdoctest_namespace) -> None:
    """Skip tests that require matplotlib if it is not installed."""

    def _is_matplotlib_installed():
        try:
            import matplotlib  # noqa: F401

        except ImportError:
            pytest.skip("This doctest requires matplotlib to be installed.")

    xdoctest_namespace["is_matplotlib_installed"] = _is_matplotlib_installed


@pytest.fixture(scope="session", autouse=True)
def doctest_setup(xdoctest_namespace, nimbus, worker_id, open_dataset) -> None:
    """Gather testing data on doctest run."""
    testing_setup_warnings()
    gather_testing_data(worker_cache_dir=nimbus.path, worker_id=worker_id)
    xdoctest_namespace.update(
        generate_atmos(branch=TESTDATA_BRANCH, cache_dir=nimbus.path)
    )

    class AttrDict(dict):
        """A dictionary that allows access to its keys as attributes."""

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.__dict__ = self

    xdoctest_namespace["open_dataset"] = open_dataset
    xdoctest_namespace["xr"] = AttrDict()
    xdoctest_namespace["xr"].update({"open_dataset": open_dataset})
    xdoctest_namespace.update(add_doctest_filepaths())
    xdoctest_namespace.update(add_example_file_paths())



"""
====================
Data files for xclim
====================

JSON and YAML definitions for virtual modules and internationalisation support.

Currently, the following virtual modules are defined:
  * ANUCLIM (`anuclim.yml`)
  * CF (`cf.yml`)
  * ICCLIM (`icclim.yml`)

And the following languages are supported:
  * English (default)
  * French (`fr.json`)

These files are used by xclim to define new indicators and to provide translations for the user interface.

Additionally, this package contains the following data files:
  * `schema.yml`: YAML schema for detailing class definitions used for indicators.
  * `variables.yml`: YAML schema defining the variables and their metadata used in the indicator definitions.
"""



from __future__ import annotations

from importlib.resources import as_file, files
from typing import NewType, TypeVar

import xarray as xr
from pint import Quantity
from yaml import safe_load

__all__ = [
    "VARIABLES",
    "DateStr",
    "DayOfYearStr",
    "Quantified",
]

#: Type annotation for strings representing full dates (YYYY-MM-DD), may include time.
DateStr = NewType("DateStr", str)

#: Type annotation for strings representing dates without a year (MM-DD).
DayOfYearStr = NewType("DayOfYearStr", str)

#: Type annotation for thresholds and other not-exactly-a-variable quantities
Quantified = TypeVar("Quantified", xr.DataArray, str, Quantity)


with as_file(files("xclim.data")) as data_dir:
    with (data_dir / "variables.yml").open() as f:
        VARIABLES = safe_load(f)["variables"]
        """Official variables definitions.

A mapping from variable name to a dict with the following keys:

- canonical_units [required] : The conventional units used by this variable.
- cell_methods [optional] : The conventional `cell_methods` CF attribute
- description [optional] : A description of the variable, to populate dynamically generated docstrings.
- dimensions [optional] : The dimensionality of the variable, an abstract version of the units.
  See `xclim.units.units._dimensions.keys()` for available terms. This is especially useful for making xclim aware of
  "[precipitation]" variables.
- standard_name [optional] : If it exists, the CF standard name.
- data_flags [optional] : Data flags methods (:py:mod:`xclim.core.dataflags`) applicable to this variable.
  The method names are keys and values are dicts of keyword arguments to pass
  (an empty dict if there's nothing to configure).
"""



"""
Generic Indices Submodule
=========================

Helper functions for common generic actions done in the computation of indices.
"""

from __future__ import annotations

import warnings
from collections.abc import Callable, Sequence

import cftime
import numpy as np
import xarray as xr
from pint import Quantity
from xarray.coding.cftime_offsets import _MONTH_ABBREVIATIONS  # noqa

from xclim.core import DayOfYearStr, Quantified
from xclim.core.calendar import doy_to_days_since, get_calendar, select_time
from xclim.core.units import (
    convert_units_to,
    declare_relative_units,
    infer_context,
    pint2cfattrs,
    pint2cfunits,
    str2pint,
    to_agg_units,
    units2pint,
)
from xclim.indices import run_length as rl
from xclim.indices.helpers import resample_map

__all__ = [
    "aggregate_between_dates",
    "binary_ops",
    "bivariate_spell_length_statistics",
    "compare",
    "count_level_crossings",
    "count_occurrences",
    "cumulative_difference",
    "default_freq",
    "detrend",
    "diurnal_temperature_range",
    "domain_count",
    "doymax",
    "doymin",
    "extreme_temperature_range",
    "first_day_threshold_reached",
    "first_occurrence",
    "get_daily_events",
    "get_op",
    "get_zones",
    "interday_diurnal_temperature_range",
    "last_occurrence",
    "season",
    "select_resample_op",
    "spell_length",
    "spell_length_statistics",
    "spell_mask",
    "statistics",
    "temperature_sum",
    "threshold_count",
    "thresholded_statistics",
]

binary_ops = {">": "gt", "<": "lt", ">=": "ge", "<=": "le", "==": "eq", "!=": "ne"}


def select_resample_op(
    da: xr.DataArray, op: str | Callable, freq: str = "YS", out_units=None, **indexer
) -> xr.DataArray:
    """Apply operation over each period that is part of the index selection.

    Parameters
    ----------
    da : xr.DataArray
        Input data.
    op : str {'min', 'max', 'mean', 'std', 'var', 'count', 'sum', 'integral', 'argmax', 'argmin'} or func
        Reduce operation. Can either be a DataArray method or a function that can be applied to a DataArray.
    freq : str
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`.
    out_units : str, optional
        Output units to assign. Only necessary if `op` is function not supported by :py:func:`xclim.core.units.to_agg_units`.
    indexer : {dim: indexer, }, optional
        Time attribute and values over which to subset the array. For example, use season='DJF' to select winter values,
        month=1 to select January, or month=[6,7,8] to select summer months. If not indexer is given, all values are
        considered.

    Returns
    -------
    xr.DataArray
        The maximum value for each period.
    """
    da = select_time(da, **indexer)
    if isinstance(op, str):
        op = _xclim_ops.get(op, op)
    if isinstance(op, str):
        out = getattr(da.resample(time=freq), op.replace("integral", "sum"))(
            dim="time", keep_attrs=True
        )
    else:
        with xr.set_options(keep_attrs=True):
            out = resample_map(da, "time", freq, op)
        op = op.__name__
    if out_units is not None:
        return out.assign_attrs(units=out_units)

    if op in ["std", "var"]:
        out.attrs.update(
            pint2cfattrs(units2pint(out.attrs["units"]), is_difference=True)
        )

    return to_agg_units(out, da, op)


def select_rolling_resample_op(
    da: xr.DataArray,
    op: str,
    window: int,
    window_center: bool = True,
    window_op: str = "mean",
    freq: str = "YS",
    out_units=None,
    **indexer,
) -> xr.DataArray:
    """Apply operation over each period that is part of the index selection, using a rolling window before the operation.

    Parameters
    ----------
    da : xr.DataArray
        Input data.
    op : str {'min', 'max', 'mean', 'std', 'var', 'count', 'sum', 'integral', 'argmax', 'argmin'} or func
        Reduce operation. Can either be a DataArray method or a function that can be applied to a DataArray.
    window : int
        Size of the rolling window (centered).
    window_center : bool
        If True, the window is centered on the date. If False, the window is right-aligned.
    window_op : str {'min', 'max', 'mean', 'std', 'var', 'count', 'sum', 'integral'}
        Operation to apply to the rolling window. Default: 'mean'.
    freq : str
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`. Applied after the rolling window.
    out_units : str, optional
        Output units to assign. Only necessary if `op` is function not supported by :py:func:`xclim.core.units.to_agg_units`.
    indexer : {dim: indexer, }, optional
        Time attribute and values over which to subset the array. For example, use season='DJF' to select winter values,
        month=1 to select January, or month=[6,7,8] to select summer months. If not indexer is given, all values are
        considered.

    Returns
    -------
    xr.DataArray
        The array for which the operation has been applied over each period.
    """
    rolled = getattr(
        da.rolling(time=window, center=window_center),
        window_op.replace("integral", "sum"),
    )()
    rolled = to_agg_units(rolled, da, window_op)
    return select_resample_op(rolled, op=op, freq=freq, out_units=out_units, **indexer)


def doymax(da: xr.DataArray) -> xr.DataArray:
    """Return the day of year of the maximum value."""
    i = da.argmax(dim="time")
    out = da.time.dt.dayofyear.isel(time=i, drop=True)
    return to_agg_units(out, da, "doymax")


def doymin(da: xr.DataArray) -> xr.DataArray:
    """Return the day of year of the minimum value."""
    i = da.argmin(dim="time")
    out = da.time.dt.dayofyear.isel(time=i, drop=True)
    return to_agg_units(out, da, "doymin")


_xclim_ops = {"doymin": doymin, "doymax": doymax}


def default_freq(**indexer) -> str:
    """Return the default frequency."""
    freq = "YS-JAN"
    if indexer:
        group, value = indexer.popitem()
        if group == "season":
            month = 12  # The "season" scheme is based on YS-DEC
        elif group == "month":
            month = np.take(value, 0)
        elif group == "doy_bounds":
            month = cftime.num2date(value[0] - 1, "days since 2004-01-01").month
        elif group == "date_bounds":
            month = int(value[0][:2])
        else:
            raise ValueError(f"Unknown group `{group}`.")
        freq = "YS-" + _MONTH_ABBREVIATIONS[month]
    return freq


def get_op(op: str, constrain: Sequence[str] | None = None) -> Callable:
    """Get python's comparing function according to its name of representation and validate allowed usage.

    Accepted op string are keys and values of xclim.indices.generic.binary_ops.

    Parameters
    ----------
    op : str
        Operator.
    constrain : sequence of str, optional
        A tuple of allowed operators.
    """
    if op == "gteq":
        warnings.warn(f"`{op}` is being renamed `ge` for compatibility.")
        op = "ge"
    if op == "lteq":
        warnings.warn(f"`{op}` is being renamed `le` for compatibility.")
        op = "le"

    if op in binary_ops:
        binary_op = binary_ops[op]
    elif op in binary_ops.values():
        binary_op = op
    else:
        raise ValueError(f"Operation `{op}` not recognized.")

    constraints = []
    if isinstance(constrain, list | tuple | set):
        constraints.extend([binary_ops[c] for c in constrain])
        constraints.extend(constrain)
    elif isinstance(constrain, str):
        constraints.extend([binary_ops[constrain], constrain])

    if constrain:
        if op not in constraints:
            raise ValueError(f"Operation `{op}` not permitted for indice.")

    return xr.core.ops.get_op(binary_op)  # noqa


def compare(
    left: xr.DataArray,
    op: str,
    right: float | int | np.ndarray | xr.DataArray,
    constrain: Sequence[str] | None = None,
) -> xr.DataArray:
    """Compare a dataArray to a threshold using given operator.

    Parameters
    ----------
    left : xr.DataArray
        A DatArray being evaluated against `right`.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}
        Logical operator. e.g. arr > thresh.
    right : float, int, np.ndarray, or xr.DataArray
        A value or array-like being evaluated against left`.
    constrain : sequence of str, optional
        Optionally allowed conditions.

    Returns
    -------
    xr.DataArray
        Boolean mask of the comparison.
    """
    return get_op(op, constrain)(left, right)


def threshold_count(
    da: xr.DataArray,
    op: str,
    threshold: float | int | xr.DataArray,
    freq: str,
    constrain: Sequence[str] | None = None,
) -> xr.DataArray:
    """Count number of days where value is above or below threshold.

    Parameters
    ----------
    da : xr.DataArray
        Input data.
    op : {">", "<", ">=", "<=", "gt", "lt", "ge", "le"}
        Logical operator. e.g. arr > thresh.
    threshold : Union[float, int]
        Threshold value.
    freq : str
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`.
    constrain : sequence of str, optional
        Optionally allowed conditions.

    Returns
    -------
    xr.DataArray
        The number of days meeting the constraints for each period.
    """
    if constrain is None:
        constrain = (">", "<", ">=", "<=")

    c = compare(da, op, threshold, constrain) * 1
    return c.resample(time=freq).sum(dim="time")


def domain_count(
    da: xr.DataArray,
    low: float | int | xr.DataArray,
    high: float | int | xr.DataArray,
    freq: str,
) -> xr.DataArray:
    """Count number of days where value is within low and high thresholds.

    A value is counted if it is larger than `low`, and smaller or equal to `high`, i.e. in `]low, high]`.

    Parameters
    ----------
    da : xr.DataArray
        Input data.
    low : scalar or DataArray
        Minimum threshold value.
    high : scalar or DataArray
        Maximum threshold value.
    freq : str
        Resampling frequency defining the periods defined in :ref:`timeseries.resampling`.

    Returns
    -------
    xr.DataArray
        The number of days where value is within [low, high] for each period.
    """
    c = compare(da, ">", low) * compare(da, "<=", high) * 1
    return c.resample(time=freq).sum(dim="time")


def get_daily_events(
    da: xr.DataArray,
    threshold: float | int | xr.DataArray,
    op: str,
    constrain: Sequence[str] | None = None,
) -> xr.DataArray:
    """Return a 0/1 mask when a condition is True or False.

    Parameters
    ----------
    da : xr.DataArray
        Input data.
    threshold : float
        Threshold value.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}
        Logical operator. e.g. arr > thresh.
    constrain : sequence of str, optional
        Optionally allowed conditions.

    Notes
    -----
    The function returns:

    - ``1`` where operator(da, da_value) is ``True``
    - ``0`` where operator(da, da_value) is ``False``
    - ``nan`` where da is ``nan``

    Returns
    -------
    xr.DataArray
    """
    events = compare(da, op, threshold, constrain) * 1
    events = events.where(~(np.isnan(da)))
    events = events.rename("events")
    return events


def spell_mask(
    data: xr.DataArray | Sequence[xr.DataArray],
    window: int,
    win_reducer: str,
    op: str,
    thresh: float | Sequence[float],
    min_gap: int = 1,
    weights: Sequence[float] = None,
    var_reducer: str = "all",
) -> xr.DataArray:
    """Compute the boolean mask of data points that are part of a spell as defined by a rolling statistic.

    A day is part of a spell (True in the mask) if it is contained in any period that fulfills the condition.

    Parameters
    ----------
    data: DataArray or sequence of DataArray
        The input data. Can be a list, in which case the condition is checked on all variables.
        See var_reducer for the latter case.
    window: int
        The length of the rolling window in which to compute statistics.
    win_reducer: {'min', 'max', 'sum', 'mean'}
        The statistics to compute on the rolling window.
    op: {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}
        The comparison operator to use when finding spells.
    thresh: float or sequence of floats
        The threshold to compare the rolling statistics against, as ``window_stats op threshold``.
        If data is a list, this must be a list of the same length with a threshold for each variable.
        This function does not handle units and can't accept Quantified objects.
    min_gap: int
        The shortest possible gap between two spells. Spells closer than this are merged by assigning
        the gap steps to the merged spell.
    weights: sequence of floats
        A list of weights of the same length as the window.
        Only supported if `win_reducer` is "mean".
    var_reducer: {'all', 'any'}
        If the data is a list, the condition must either be fulfilled on *all*
        or *any* variables for the period to be considered a spell.

    Returns
    -------
    xr.DataArray
        Same shape as ``data``, but boolean.
        If ``data`` was a list, this is a DataArray of the same shape as the alignment of all variables.
    """
    # Checks
    if not isinstance(data, xr.DataArray):
        # thus a sequence
        if np.isscalar(thresh) or len(data) != len(thresh):
            raise ValueError(
                "When ``data`` is given as a list, ``threshold`` must be a sequence of the same length."
            )
        data = xr.concat(data, "variable")
        if isinstance(thresh[0], xr.DataArray):
            thresh = xr.concat(thresh, "variable")
        else:
            thresh = xr.DataArray(thresh, dims=("variable",))
    if weights is not None:
        if win_reducer != "mean":
            raise ValueError(
                f"Argument 'weights' is only supported if 'win_reducer' is 'mean'. Got :  {win_reducer}"
            )
        elif len(weights) != window:
            raise ValueError(
                f"Weights have a different length ({len(weights)}) than the window ({window})."
            )
        weights = xr.DataArray(weights, dims=("window",))

    if window == 1:  # Fast path
        is_in_spell = compare(data, op, thresh)
        if not np.isscalar(thresh):
            is_in_spell = getattr(is_in_spell, var_reducer)("variable")
    elif (win_reducer == "min" and op in [">", ">=", "ge", "gt"]) or (
        win_reducer == "max" and op in ["`<", "<=", "le", "lt"]
    ):
        # Fast path for specific cases, this yields a smaller dask graph (rolling twice is expensive!)
        # For these two cases, a day can't be part of a spell if it doesn't respect the condition itself
        mask = compare(data, op, thresh)
        if not np.isscalar(thresh):
            mask = getattr(mask, var_reducer)("variable")
        # We need to filter out the spells shorter than "window"
        # find sequences of consecutive respected constraints
        cs_s = rl._cumsum_reset(mask)
        # end of these sequences
        cs_s = cs_s.where(mask.shift({"time": -1}, fill_value=0) == 0)
        # propagate these end of sequences
        # the `.where(mask>0, 0)` acts a stopper
        is_in_spell = cs_s.where(cs_s >= window).where(mask > 0, 0).bfill("time") > 0
    else:
        data_pad = data.pad(time=(0, window))
        # The spell-wise value to test
        # For example "window_reducer='sum'", we want the sum over the minimum spell length (window) to be above the thresh
        if weights is not None:
            spell_value = data_pad.rolling(time=window).construct("window").dot(weights)
        else:
            spell_value = getattr(data_pad.rolling(time=window), win_reducer)()
        # True at the end of a spell respecting the condition
        mask = compare(spell_value, op, thresh)
        if not np.isscalar(thresh):
            mask = getattr(mask, var_reducer)("variable")
        # True for all days part of a spell that respected the condition (shift because of the two rollings)
        is_in_spell = (mask.rolling(time=window).sum() >= 1).shift(
            time=-(window - 1), fill_value=False
        )
        # Cut back to the original size
        is_in_spell = is_in_spell.isel(time=slice(0, data.time.size))

    if min_gap > 1:
        is_in_spell = rl.runs_with_holes(is_in_spell, 1, ~is_in_spell, min_gap).astype(
            bool
        )

    return is_in_spell


def _spell_length_statistics(
    data: xr.DataArray | Sequence[xr.DataArray],
    thresh: float | xr.DataArray | Sequence[xr.DataArray] | Sequence[float],
    window: int,
    win_reducer: str,
    op: str,
    spell_reducer: str | Sequence[str],
    freq: str,
    min_gap: int = 1,
    resample_before_rl: bool = True,
    **indexer,
) -> xr.DataArray | Sequence[xr.DataArray]:
    if isinstance(spell_reducer, str):
        spell_reducer = [spell_reducer]
    is_in_spell = spell_mask(
        data, window, win_reducer, op, thresh, min_gap=min_gap
    ).astype(np.float32)
    is_in_spell = select_time(is_in_spell, **indexer)

    outs = []
    for sr in spell_reducer:
        out = rl.resample_and_rl(
            is_in_spell,
            resample_before_rl,
            rl.rle_statistics,
            reducer=sr,
            # The code above already ensured only spell of the minimum length are selected
            window=1,
            freq=freq,
        )

        if sr == "count":
            outs.append(out.assign_attrs(units=""))
        else:
            # All other cases are statistics of the number of timesteps
            outs.append(
                to_agg_units(
                    out,
                    data if isinstance(data, xr.DataArray) else data[0],
                    "count",
                )
            )
    if len(outs) == 1:
        return outs[0]
    return tuple(outs)


@declare_relative_units(threshold="<data>")
def spell_length_statistics(
    data: xr.DataArray,
    threshold: Quantified,
    window: int,
    win_reducer: str,
    op: str,
    spell_reducer: str,
    freq: str,
    min_gap: int = 1,
    resample_before_rl: bool = True,
    **indexer,
):
    r"""Statistics on spells lengths.

    A spell is when a statistic (`win_reducer`) over a minimum number (`window`) of consecutive timesteps respects a condition (`op` `thresh`).
    This returns a statistic over the spells count or lengths.

    Parameters
    ----------
    data : xr.DataArray
        Input data.
    threshold : Quantified
        Threshold to test against.
    window : int
        Minimum length of a spell.
    win_reducer : {'min', 'max', 'sum', 'mean'}
        Reduction along the spell length to compute the spell value.
        Note that this does not matter when `window` is 1.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}
        Logical operator. Ex: spell_value > thresh.
    spell_reducer : {'max', 'sum', 'count'} or sequence thereof
        Statistic on the spell lengths. If a list, multiple statistics are computed.
    freq : str
        Resampling frequency.
    min_gap : int
        The shortest possible gap between two spells. Spells closer than this are merged by assigning
        the gap steps to the merged spell.
    resample_before_rl : bool
        Determines if the resampling should take place before or after the run
        length encoding (or a similar algorithm) is applied to runs.
    \*\*indexer
        Indexing parameters to compute the indicator on a temporal subset of the data.
        It accepts the same arguments as :py:func:`xclim.indices.generic.select_time`.
        Indexing is done after finding the days part of a spell, but before taking the spell statistics.

    Examples
    --------
    >>> spell_length_statistics(
    ...     tas,
    ...     threshold="35 °C",
    ...     window=7,
    ...     op=">",
    ...     win_reducer="min",
    ...     spell_reducer="sum",
    ...     freq="YS",
    ... )

    Here, a day is part of a spell if it is in any seven (7) day period where the minimum temperature is over 35°C.
    We then return the annual sum of the spell lengths, so the total number of days in such spells.
    >>> from xclim.core.units import rate2amount
    >>> pram = rate2amount(pr, out_units="mm")
    >>> spell_length_statistics(
    ...     pram,
    ...     threshold="20 mm",
    ...     window=5,
    ...     op=">=",
    ...     win_reducer="sum",
    ...     spell_reducer="max",
    ...     freq="YS",
    ... )

    Here, a day is part of a spell if it is in any five (5) day period where the total accumulated precipitation reaches
    or exceeds 20 mm. We then return the length of the longest of such spells.

    See Also
    --------
    spell_mask : The lower level functions that finds spells.
    bivariate_spell_length_statistics : The bivariate version of this function.
    """
    thresh = convert_units_to(threshold, data, context="infer")
    return _spell_length_statistics(
        data,
        thresh,
        window,
        win_reducer,
        op,
        spell_reducer,
        freq,
        min_gap=min_gap,
        resample_before_rl=resample_before_rl,
        **indexer,
    )


@declare_relative_units(threshold1="<data1>", threshold2="<data2>")
def bivariate_spell_length_statistics(
    data1: xr.DataArray,
    threshold1: Quantified,
    data2: xr.DataArray,
    threshold2: Quantified,
    window: int,
    win_reducer: str,
    op: str,
    spell_reducer: str,
    freq: str,
    min_gap: int = 1,
    resample_before_rl: bool = True,
    **indexer,
):
    r"""Statistics on spells lengths based on two variables.

    A spell is when a statistic (`win_reducer`) over a minimum number (`window`) of consecutive timesteps respects a condition (`op` `thresh`).
    This returns a statistic over the spells count or lengths. In this bivariate version, conditions on both variables must be fulfilled.

    Parameters
    ----------
    data1 : xr.DataArray
        First input data.
    threshold1 : Quantified
        Threshold to test against data1.
    data2 : xr.DataArray
        Second input data.
    threshold2 : Quantified
        Threshold to test against data2.
    window : int
        Minimum length of a spell.
    win_reducer : {'min', 'max', 'sum', 'mean'}
        Reduction along the spell length to compute the spell value.
        Note that this does not matter when `window` is 1.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}
        Logical operator. Ex: spell_value > thresh.
    spell_reducer : {'max', 'sum', 'count'} or sequence thereof
        Statistic on the spell lengths. If a list, multiple statistics are computed.
    freq : str
        Resampling frequency.
    min_gap : int
        The shortest possible gap between two spells. Spells closer than this are merged by assigning
        the gap steps to the merged spell.
    resample_before_rl : bool
        Determines if the resampling should take place before or after the run
        length encoding (or a similar algorithm) is applied to runs.
    \*\*indexer
        Indexing parameters to compute the indicator on a temporal subset of the data.
        It accepts the same arguments as :py:func:`xclim.indices.generic.select_time`.
        Indexing is done after finding the days part of a spell, but before taking the spell statistics.

    See Also
    --------
    spell_length_statistics: The univariate version.
    spell_mask : The lower level functions that finds spells.
    """
    thresh1 = convert_units_to(threshold1, data1, context="infer")
    thresh2 = convert_units_to(threshold2, data2, context="infer")
    return _spell_length_statistics(
        [data1, data2],
        [thresh1, thresh2],
        window,
        win_reducer,
        op,
        spell_reducer,
        freq,
        min_gap,
        resample_before_rl,
        **indexer,
    )


@declare_relative_units(thresh="<data>")
def season(
    data: xr.DataArray,
    thresh: Quantified,
    window: int,
    op: str,
    stat: str,
    freq: str,
    mid_date: DayOfYearStr | None = None,
    constrain: Sequence[str] | None = None,
) -> xr.DataArray:
    r"""Season.

    A season starts when a variable respects some condition for a consecutive run of `N` days. It stops
    when the condition is inverted for `N` days. Runs where the condition is not met for fewer than `N` days
    are thus allowed. Additionally a middle date can serve as a maximal start date and minimum end date.

    Parameters
    ----------
    data : xr.DataArray
        Variable.
    thresh : Quantified
        Threshold on which to base evaluation.
    window : int
        Minimum number of days that the condition must be met / not met for the start / end of the season.
    op : str
        Comparison operation.
    stat : {'start', 'end', 'length'}
        Which season facet to return.
    freq : str
        Resampling frequency.
    mid_date : DayOfYearStr, optional
        An optional middle date. The start must happen before and the end after for the season to be valid.
    constrain : Sequence of strings, optional
        A list of acceptable comparison operators. Optional, but indicators wrapping this function should inject it.

    Returns
    -------
    xr.DataArray, [dimensionless] or [time]
        Depends on 'stat'. If 'start' or 'end', this is the day of year of the season's start or end.
        If 'length', this is the length of the season.

    Examples
    --------
    >>> season(tas, thresh="0 °C", window=5, op=">", stat="start", freq="YS")

    Returns the start of the "frost-free" season. The season starts with 5 consecutive days with mean temperature
    above 0°C and ends with as many days under or equal to 0°C. And end does not need to be found for a start to be valid.

    >>> season(
    ...     pr,
    ...     thresh="2 mm/d",
    ...     window=7,
    ...     op="<=",
    ...     mid_date="08-01",
    ...     stat="length",
    ...     freq="YS",
    ... )

    Returns the length of the "dry" season. The season starts with 7 consecutive days with precipitation under or equal to
    2 mm/d and ends with as many days above 2 mm/d. If no start is found before the first of august, the season is invalid.
    If a start is found but no end, the end is set to the last day of the period (December 31st if the dataset is complete).

    See Also
    --------
    xclim.indices.run_length.season_start
    xclim.indices.run_length.season_length
    xclim.indices.run_length.season_end
    """
    thresh = convert_units_to(thresh, data, context="infer")
    cond = compare(data, op, thresh, constrain=constrain)
    FUNC = {"start": rl.season_start, "end": rl.season_end, "length": rl.season_length}
    map_kwargs = {"window": window, "mid_date": mid_date}
    if stat in ["start", "end"]:
        map_kwargs["coord"] = "dayofyear"
    out = resample_map(cond, "time", freq, FUNC[stat], map_kwargs=map_kwargs)
    if stat == "length":
        return to_agg_units(out, data, "count")
    # else, a date
    out.attrs.update(units="", is_dayofyear=np.int32(1), calendar=get_calendar(data))
    return out


# CF-INDEX-META Indices


@declare_relative_units(threshold="<low_data>")
def count_level_crossings(
    low_data: xr.DataArray,
    high_data: xr.DataArray,
    threshold: Quantified,
    freq: str,
    *,
    op_low: str = "<",
    op_high: str = ">=",
) -> xr.DataArray:
    """Calculate the number of times low_data is below threshold while high_data is above threshold.

    First, the threshold is transformed to the same standard_name and units as the input data,
    then the thresholding is performed, and finally, the number of occurrences is counted.

    Parameters
    ----------
    low_data : xr.DataArray
        Variable that must be under the threshold.
    high_data : xr.DataArray
        Variable that must be above the threshold.
    threshold : Quantified
        Threshold.
    freq : str
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`.
    op_low : {"<", "<=", "lt", "le"}
        Comparison operator for low_data. Default: "<".
    op_high : {">", ">=", "gt", "ge"}
        Comparison operator for high_data. Default: ">=".

    Returns
    -------
    xr.DataArray
    """
    # Convert units to low_data
    high_data = convert_units_to(high_data, low_data)
    threshold = convert_units_to(threshold, low_data)

    lower = compare(low_data, op_low, threshold, constrain=("<", "<="))
    higher = compare(high_data, op_high, threshold, constrain=(">", ">="))

    out = (lower & higher).resample(time=freq).sum()
    return to_agg_units(out, low_data, "count", dim="time")


@declare_relative_units(threshold="<data>")
def count_occurrences(
    data: xr.DataArray,
    threshold: Quantified,
    freq: str,
    op: str,
    constrain: Sequence[str] | None = None,
) -> xr.DataArray:
    """Calculate the number of times some condition is met.

    First, the threshold is transformed to the same standard_name and units as the input data.
    Then the thresholding is performed as condition(data, threshold),
    i.e. if condition is `<`, then this counts the number of times `data < threshold`.
    Finally, count the number of occurrences when condition is met.

    Parameters
    ----------
    data : xr.DataArray
        An array.
    threshold : Quantified
        Threshold.
    freq : str
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}
        Logical operator. e.g. arr > thresh.
    constrain : sequence of str, optional
        Optionally allowed conditions.

    Returns
    -------
    xr.DataArray
    """
    threshold = convert_units_to(threshold, data)

    cond = compare(data, op, threshold, constrain)

    out = cond.resample(time=freq).sum()
    return to_agg_units(out, data, "count", dim="time")


def diurnal_temperature_range(
    low_data: xr.DataArray, high_data: xr.DataArray, reducer: str, freq: str
) -> xr.DataArray:
    """Calculate the diurnal temperature range and reduce according to a statistic.

    Parameters
    ----------
    low_data : xr.DataArray
        The lowest daily temperature (tasmin).
    high_data : xr.DataArray
        The highest daily temperature (tasmax).
    reducer : {'max', 'min', 'mean', 'sum'}
        Reducer.
    freq: str
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`.

    Returns
    -------
    xr.DataArray
    """
    high_data = convert_units_to(high_data, low_data)

    dtr = high_data - low_data
    out = getattr(dtr.resample(time=freq), reducer)()

    u = str2pint(low_data.units)
    out.attrs.update(pint2cfattrs(u, is_difference=True))
    return out


@declare_relative_units(threshold="<data>")
def first_occurrence(
    data: xr.DataArray,
    threshold: Quantified,
    freq: str,
    op: str,
    constrain: Sequence[str] | None = None,
) -> xr.DataArray:
    """Calculate the first time some condition is met.

    First, the threshold is transformed to the same standard_name and units as the input data.
    Then the thresholding is performed as condition(data, threshold), i.e. if condition is <, data < threshold.
    Finally, locate the first occurrence when condition is met.

    Parameters
    ----------
    data : xr.DataArray
        Input data.
    threshold : Quantified
        Threshold.
    freq : str
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}
        Logical operator. e.g. arr > thresh.
    constrain : sequence of str, optional
        Optionally allowed conditions.

    Returns
    -------
    xr.DataArray
    """
    threshold = convert_units_to(threshold, data)

    cond = compare(data, op, threshold, constrain)

    out = resample_map(
        cond,
        "time",
        freq,
        rl.first_run,
        map_kwargs=dict(window=1, dim="time", coord="dayofyear"),
    )
    out.attrs["units"] = ""
    return out


@declare_relative_units(threshold="<data>")
def last_occurrence(
    data: xr.DataArray,
    threshold: Quantified,
    freq: str,
    op: str,
    constrain: Sequence[str] | None = None,
) -> xr.DataArray:
    """Calculate the last time some condition is met.

    First, the threshold is transformed to the same standard_name and units as the input data.
    Then the thresholding is performed as condition(data, threshold), i.e. if condition is <, data < threshold.
    Finally, locate the last occurrence when condition is met.

    Parameters
    ----------
    data : xr.DataArray
        Input data.
    threshold : Quantified
        Threshold.
    freq : str
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}
        Logical operator. e.g. arr > thresh.
    constrain : sequence of str, optional
        Optionally allowed conditions.

    Returns
    -------
    xr.DataArray
    """
    threshold = convert_units_to(threshold, data)

    cond = compare(data, op, threshold, constrain)

    out = resample_map(
        cond,
        "time",
        freq,
        rl.last_run,
        map_kwargs=dict(window=1, dim="time", coord="dayofyear"),
    )
    out.attrs["units"] = ""
    return out


@declare_relative_units(threshold="<data>")
def spell_length(
    data: xr.DataArray, threshold: Quantified, reducer: str, freq: str, op: str
) -> xr.DataArray:
    """Calculate statistics on lengths of spells.

    First, the threshold is transformed to the same standard_name and units as the input data.
    Then the thresholding is performed as condition(data, threshold), i.e. if condition is <, data < threshold.
    Then the spells are determined, and finally the statistics according to the specified reducer are calculated.

    Parameters
    ----------
    data : xr.DataArray
        Input data.
    threshold : Quantified
        Threshold.
    reducer : {'max', 'min', 'mean', 'sum'}
        Reducer.
    freq : str
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}
        Logical operator. e.g. arr > thresh.

    Returns
    -------
    xr.DataArray
    """
    threshold = convert_units_to(
        threshold,
        data,
        context=infer_context(standard_name=data.attrs.get("standard_name")),
    )

    cond = compare(data, op, threshold)

    out = resample_map(
        cond,
        "time",
        freq,
        rl.rle_statistics,
        map_kwargs=dict(reducer=reducer, window=1, dim="time"),
    )
    return to_agg_units(out, data, "count")


def statistics(data: xr.DataArray, reducer: str, freq: str) -> xr.DataArray:
    """Calculate a simple statistic of the data.

    Parameters
    ----------
    data : xr.DataArray
        Input data.
    reducer : {'max', 'min', 'mean', 'sum'}
        Reducer.
    freq : str
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`.

    Returns
    -------
    xr.DataArray
    """
    out = getattr(data.resample(time=freq), reducer)()
    out.attrs["units"] = data.attrs["units"]
    return out


@declare_relative_units(threshold="<data>")
def thresholded_statistics(
    data: xr.DataArray,
    op: str,
    threshold: Quantified,
    reducer: str,
    freq: str,
    constrain: Sequence[str] | None = None,
) -> xr.DataArray:
    """Calculate a simple statistic of the data for which some condition is met.

    First, the threshold is transformed to the same standard_name and units as the input data.
    Then the thresholding is performed as condition(data, threshold), i.e. if condition is <, data < threshold.
    Finally, the statistic is calculated for those data values that fulfill the condition.

    Parameters
    ----------
    data : xr.DataArray
        Input data.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}
        Logical operator. e.g. arr > thresh.
    threshold : Quantified
        Threshold.
    reducer : {'max', 'min', 'mean', 'sum'}
        Reducer.
    freq : str
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`.
    constrain : sequence of str, optional
        Optionally allowed conditions. Default: None.

    Returns
    -------
    xr.DataArray
    """
    threshold = convert_units_to(threshold, data)

    cond = compare(data, op, threshold, constrain)

    out = getattr(data.where(cond).resample(time=freq), reducer)()
    out.attrs["units"] = data.attrs["units"]
    return out


@declare_relative_units(threshold="<data>")
def temperature_sum(
    data: xr.DataArray, op: str, threshold: Quantified, freq: str
) -> xr.DataArray:
    """Calculate the temperature sum above/below a threshold.

    First, the threshold is transformed to the same standard_name and units as the input data.
    Then the thresholding is performed as condition(data, threshold), i.e. if condition is <, data < threshold.
    Finally, the sum is calculated for those data values that fulfill the condition after subtraction of the threshold
    value. If the sum is for values below the threshold the result is multiplied by -1.

    Parameters
    ----------
    data : xr.DataArray
        Input data.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le"}
        Logical operator. e.g. arr > thresh.
    threshold : Quantified
        Threshold.
    freq : str
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`.

    Returns
    -------
    xr.DataArray
    """
    threshold = convert_units_to(threshold, data)

    cond = compare(data, op, threshold, constrain=("<", "<=", ">", ">="))
    direction = -1 if op in ["<", "<=", "lt", "le"] else 1

    out = (data - threshold).where(cond).resample(time=freq).sum()
    out = direction * out
    out.attrs["units_metadata"] = "temperature: difference"
    return to_agg_units(out, data, "integral")


def interday_diurnal_temperature_range(
    low_data: xr.DataArray, high_data: xr.DataArray, freq: str
) -> xr.DataArray:
    """Calculate the average absolute day-to-day difference in diurnal temperature range.

    Parameters
    ----------
    low_data : xr.DataArray
        The lowest daily temperature (tasmin).
    high_data : xr.DataArray
        The highest daily temperature (tasmax).
    freq : str
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`.

    Returns
    -------
    xr.DataArray
    """
    high_data = convert_units_to(high_data, low_data)

    vdtr = abs((high_data - low_data).diff(dim="time"))
    out = vdtr.resample(time=freq).mean(dim="time")

    out.attrs["units"] = low_data.attrs["units"]
    out.attrs["units_metadata"] = "temperature: difference"
    return out


def extreme_temperature_range(
    low_data: xr.DataArray, high_data: xr.DataArray, freq: str
) -> xr.DataArray:
    """Calculate the extreme temperature range as the maximum of daily maximum temperature minus the minimum of daily minimum temperature.

    Parameters
    ----------
    low_data : xr.DataArray
        The lowest daily temperature (tasmin).
    high_data : xr.DataArray
        The highest daily temperature (tasmax).
    freq : str
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`.

    Returns
    -------
    xr.DataArray
    """
    high_data = convert_units_to(high_data, low_data)

    out = high_data.resample(time=freq).max() - low_data.resample(time=freq).min()

    out.attrs["units"] = low_data.attrs["units"]
    out.attrs["units_metadata"] = "temperature: difference"
    return out


def aggregate_between_dates(
    data: xr.DataArray,
    start: xr.DataArray | DayOfYearStr,
    end: xr.DataArray | DayOfYearStr,
    op: str = "sum",
    freq: str | None = None,
) -> xr.DataArray:
    """Aggregate the data over a period between start and end dates and apply the operator on the aggregated data.

    Parameters
    ----------
    data : xr.DataArray
        Data to aggregate between start and end dates.
    start : xr.DataArray or DayOfYearStr
        Start dates (as day-of-year) for the aggregation periods.
    end : xr.DataArray or DayOfYearStr
        End (as day-of-year) dates for the aggregation periods.
    op : {'min', 'max', 'sum', 'mean', 'std'}
        Operator.
    freq : str, optional
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`. Default: `None`.

    Returns
    -------
    xr.DataArray, [dimensionless]
        Aggregated data between the start and end dates. If the end date is before the start date, returns np.nan.
        If there is no start and/or end date, returns np.nan.
    """

    def _get_days(_bound, _group, _base_time):
        """Get bound in number of days since base_time. Bound can be a days_since array or a DayOfYearStr."""
        if isinstance(_bound, str):
            b_i = rl.index_of_date(_group.time, _bound, max_idxs=1)  # noqa
            if not b_i.size > 0:
                return None
            return (_group.time.isel(time=b_i[0]) - _group.time.isel(time=0)).dt.days
        if _base_time in _bound.time:
            return _bound.sel(time=_base_time)
        return None

    if freq is None:
        frequencies = []
        for bound in [start, end]:
            try:
                frequencies.append(xr.infer_freq(bound.time))
            except AttributeError:
                frequencies.append(None)

        good_freq = set(frequencies) - {None}

        if len(good_freq) != 1:
            raise ValueError(
                f"Non-inferrable resampling frequency or inconsistent frequencies. Got start, end = {frequencies}."
                " Please consider providing `freq` manually."
            )
        freq = good_freq.pop()

    cal = get_calendar(data, dim="time")

    if not isinstance(start, str):
        start = start.convert_calendar(cal)
        start.attrs["calendar"] = cal
        start = doy_to_days_since(start)
    if not isinstance(end, str):
        end = end.convert_calendar(cal)
        end.attrs["calendar"] = cal
        end = doy_to_days_since(end)

    out = []
    for base_time, indexes in data.resample(time=freq).groups.items():
        # get group slice
        group = data.isel(time=indexes)

        start_d = _get_days(start, group, base_time)
        end_d = _get_days(end, group, base_time)

        # convert bounds for this group
        if start_d is not None and end_d is not None:
            days = (group.time - base_time).dt.days
            days[days < 0] = np.nan

            masked = group.where((days >= start_d) & (days <= end_d - 1))
            res = getattr(masked, op)(dim="time", skipna=True)
            res = xr.where(
                ((start_d > end_d) | (start_d.isnull()) | (end_d.isnull())), np.nan, res
            )
            # Re-add the time dimension with the period's base time.
            res = res.expand_dims(time=[base_time])
            out.append(res)
        else:
            # Get an array with the good shape, put nans and add the new time.
            res = (group.isel(time=0) * np.nan).expand_dims(time=[base_time])
            out.append(res)
            continue

    return xr.concat(out, dim="time")


@declare_relative_units(threshold="<data>")
def cumulative_difference(
    data: xr.DataArray, threshold: Quantified, op: str, freq: str | None = None
) -> xr.DataArray:
    """Calculate the cumulative difference below/above a given value threshold.

    Parameters
    ----------
    data : xr.DataArray
        Data for which to determine the cumulative difference.
    threshold : Quantified
        The value threshold.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le"}
        Logical operator. e.g. arr > thresh.
    freq : str, optional
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`.
        If `None`, no resampling is performed. Default: `None`.

    Returns
    -------
    xr.DataArray
    """
    threshold = convert_units_to(threshold, data)

    if op in ["<", "<=", "lt", "le"]:
        diff = (threshold - data).clip(0)
    elif op in [">", ">=", "gt", "ge"]:
        diff = (data - threshold).clip(0)
    else:
        raise NotImplementedError(f"Condition not supported: '{op}'.")

    if freq is not None:
        diff = diff.resample(time=freq).sum(dim="time")

    diff.attrs.update(pint2cfattrs(units2pint(data.attrs["units"]), is_difference=True))

    return to_agg_units(diff, data, op="integral")


@declare_relative_units(threshold="<data>")
def first_day_threshold_reached(
    data: xr.DataArray,
    *,
    threshold: Quantified,
    op: str,
    after_date: DayOfYearStr,
    window: int = 1,
    freq: str = "YS",
    constrain: Sequence[str] | None = None,
) -> xr.DataArray:
    r"""First day of values exceeding threshold.

    Returns first day of period where values reach or exceed a threshold over a given number of days,
    limited to a starting calendar date.

    Parameters
    ----------
    data xr.DataArray
        Dataset being evaluated.
    threshold : str
        Threshold on which to base evaluation.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}
        Logical operator. e.g. arr > thresh.
    after_date : str
        Date of the year after which to look for the first event. Should have the format '%m-%d'.
    window : int
        Minimum number of days with values above threshold needed for evaluation. Default: 1.
    freq : str
        Resampling frequency defining the periods as defined in :ref:`timeseries.resampling`.
        Default: "YS".
    constrain : sequence of str, optional
        Optionally allowed conditions.

    Returns
    -------
    xr.DataArray, [dimensionless]
        Day of the year when value reaches or exceeds a threshold over a given number of days for the first time.
        If there is no such day, returns np.nan.
    """
    threshold = convert_units_to(threshold, data)

    cond = compare(data, op, threshold, constrain=constrain)

    out: xr.DataArray = resample_map(
        cond,
        "time",
        freq,
        rl.first_run_after_date,
        map_kwargs=dict(window=window, date=after_date, dim="time", coord="dayofyear"),
    )
    out.attrs.update(units="", is_dayofyear=np.int32(1), calendar=get_calendar(data))
    return out


def _get_zone_bins(
    zone_min: Quantity,
    zone_max: Quantity,
    zone_step: Quantity,
):
    """Bin boundary values as defined by zone parameters.

    Parameters
    ----------
    zone_min : Quantity
        Left boundary of the first zone
    zone_max : Quantity
        Right boundary of the last zone
    zone_step: Quantity
        Size of zones

    Returns
    -------
    xr.DataArray, [units of `zone_step`]
        Array of values corresponding to each zone: [zone_min, zone_min+step, ..., zone_max]
    """
    units = pint2cfunits(str2pint(zone_step))
    mn, mx, step = (
        convert_units_to(str2pint(z), units) for z in [zone_min, zone_max, zone_step]
    )
    bins = np.arange(mn, mx + step, step)
    if (mx - mn) % step != 0:
        warnings.warn(
            "`zone_max` - `zone_min` is not an integer multiple of `zone_step`. Last zone will be smaller."
        )
        bins[-1] = mx
    return xr.DataArray(bins, attrs={"units": units})


def get_zones(
    da: xr.DataArray,
    zone_min: Quantity | None = None,
    zone_max: Quantity | None = None,
    zone_step: Quantity | None = None,
    bins: xr.DataArray | list[Quantity] | None = None,
    exclude_boundary_zones: bool = True,
    close_last_zone_right_boundary: bool = True,
) -> xr.DataArray:
    r"""Divide data into zones and attribute a zone coordinate to each input value.

    Divide values into zones corresponding to bins of width zone_step beginning at zone_min and ending at zone_max.
    Bins are inclusive on the left values and exclusive on the right values.

    Parameters
    ----------
    da : xr.DataArray
        Input data
    zone_min : Quantity | None
        Left boundary of the first zone
    zone_max : Quantity | None
        Right boundary of the last zone
    zone_step: Quantity | None
        Size of zones
    bins : xr.DataArray | list[Quantity] | None
        Zones to be used, either as a DataArray with appropriate units or a list of Quantity
    exclude_boundary_zones : Bool
        Determines whether a zone value is attributed for values in ]`-np.inf`, `zone_min`[ and [`zone_max`, `np.inf`\ [.
    close_last_zone_right_boundary : Bool
        Determines if the right boundary of the last zone is closed.

    Returns
    -------
    xr.DataArray, [dimensionless]
        Zone index for each value in `da`. Zones are returned as an integer range, starting from `0`
    """
    # Check compatibility of arguments
    zone_params = np.array([zone_min, zone_max, zone_step])
    if bins is None:
        if (zone_params == [None] * len(zone_params)).any():
            raise ValueError(
                "`bins` is `None` as well as some or all of [`zone_min`, `zone_max`, `zone_step`]. "
                "Expected defined parameters in one of these cases."
            )
    elif set(zone_params) != {None}:
        warnings.warn(
            "Expected either `bins` or [`zone_min`, `zone_max`, `zone_step`], got both. "
            "`bins` will be used."
        )

    # Get zone bins (if necessary)
    bins = bins if bins is not None else _get_zone_bins(zone_min, zone_max, zone_step)
    if isinstance(bins, list):
        bins = sorted([convert_units_to(b, da) for b in bins])
    else:
        bins = convert_units_to(bins, da)

    def _get_zone(_da):
        return np.digitize(_da, bins) - 1

    zones = xr.apply_ufunc(_get_zone, da, dask="parallelized")

    if close_last_zone_right_boundary:
        zones = zones.where(da != bins[-1], _get_zone(bins[-2]))
    if exclude_boundary_zones:
        zones = zones.where(
            (zones != _get_zone(bins[0] - 1)) & (zones != _get_zone(bins[-1]))
        )

    return zones


def detrend(
    ds: xr.DataArray | xr.Dataset, dim="time", deg=1
) -> xr.DataArray | xr.Dataset:
    """Detrend data along a given dimension computing a polynomial trend of a given order.

    Parameters
    ----------
    ds : xr.Dataset or xr.DataArray
      The data to detrend. If a Dataset, detrending is done on all data variables.
    dim : str
      Dimension along which to compute the trend.
    deg : int
      Degree of the polynomial to fit.

    Returns
    -------
    xr.Dataset or xr.DataArray
      Same as `ds`, but with its trend removed (subtracted).
    """
    if isinstance(ds, xr.Dataset):
        return ds.map(detrend, keep_attrs=False, dim=dim, deg=deg)
    # is a DataArray
    # detrend along a single dimension
    coeff = ds.polyfit(dim=dim, deg=deg)
    trend = xr.polyval(ds[dim], coeff.polyfit_coefficients)
    with xr.set_options(keep_attrs=True):
        return ds - trend


@declare_relative_units(thresh="<data>")
def thresholded_events(
    data: xr.DataArray,
    thresh: Quantified,
    op: str,
    window: int,
    thresh_stop: Quantified | None = None,
    op_stop: str | None = None,
    window_stop: int = 1,
    freq: str | None = None,
) -> xr.Dataset:
    r"""Thresholded events.

    Finds all events along the time dimension. An event starts if the start condition is fulfilled for a given number of consecutive time steps.
    It ends when the end condition is fulfilled for a given number of consecutive time steps.

    Conditions are simple comparison of the data with a threshold: ``cond = data op thresh``.
    The end conditions defaults to the negation of the start condition.

    The resulting ``event`` dimension always has its maximal possible size : ``data.size / (window + window_stop)``.

    Parameters
    ----------
    data : xr.DataArray
        Variable.
    thresh : Quantified
        Threshold defining the event.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}
        Logical operator defining the event, e.g. arr > thresh.
    window: int
        Number of time steps where the event condition must be true to start an event.
    thresh_stop : Quantified, optional
        Threshold defining the end of an event. Defaults to `thresh`.
    op_stop : {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}, optional
        Logical operator for the end of an event. Defaults to the opposite of `op`.
    window_stop: int, optional
        Number of time steps where the end condition must be true to end an event. Defaults to 1.
    freq: str, optional
        A frequency to divide the data into periods. If absent, the output has not time dimension.
        If given, the events are searched within in each resample period independently.

    Returns
    -------
    xr.Dataset, same shape as the data except the time dimension is replaced by an "event" dimension.
        event_length: The number of time steps in each event
        event_effective_length: The number of time steps of even event where the start condition is true.
        event_sum: The sum within each event, only considering the steps where start condition is true.
        event_start: The datetime of the start of the run.
    """
    thresh = convert_units_to(thresh, data)

    # Start and end conditions
    da_start = compare(data, op, thresh)
    if thresh_stop is None and op_stop is None:
        da_stop = ~da_start
    else:
        thresh_stop = convert_units_to(thresh_stop or thresh, data)
        if op_stop is not None:
            da_stop = compare(data, op_stop, thresh_stop)
        else:
            da_stop = ~compare(data, op, thresh_stop)

    return rl.find_events(da_start, window, da_stop, window_stop, data, freq)



from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

from xclim.sdba import nbutils as nbu


class TestQuantiles:
    @pytest.mark.parametrize("uses_dask", [True, False])
    def test_quantile(self, open_dataset, uses_dask):
        da = (
            open_dataset("sdba/CanESM2_1950-2100.nc").sel(time=slice("1950", "1955")).pr
        ).load()
        if uses_dask:
            da = da.chunk({"location": 1})
        else:
            da = da.load()
        q = np.linspace(0.1, 0.99, 50)
        out_nbu = nbu.quantile(da, q, dim="time").transpose("location", ...)
        out_xr = da.quantile(q=q, dim="time").transpose("location", ...)
        np.testing.assert_array_almost_equal(out_nbu.values, out_xr.values)

    def test_edge_cases(self, open_dataset):
        q = np.linspace(0.1, 0.99, 50)

        # only 1 non-null value
        da = xr.DataArray([1] + [np.nan] * 100, dims="dim_0")
        out_nbu = nbu.quantile(da, q, dim="dim_0")
        np.testing.assert_array_equal(out_nbu.values, np.full_like(q, 1))

        # only NANs
        da = xr.DataArray([np.nan] * 100, dims="dim_0")
        out_nbu = nbu.quantile(da, q, dim="dim_0")
        np.testing.assert_array_equal(out_nbu.values, np.full_like(q, np.nan))



"""
Testing and Tutorial Utilities' Module
======================================
"""

from __future__ import annotations

import importlib.resources as ilr
import logging
import os
import platform
import re
import sys
import time
import warnings
from collections.abc import Sequence
from datetime import datetime as dt
from importlib import import_module
from io import StringIO
from pathlib import Path
from shutil import copytree
from typing import TextIO
from urllib.error import HTTPError, URLError
from urllib.parse import urljoin, urlparse
from urllib.request import urlretrieve

from filelock import FileLock
from packaging.version import Version
from xarray import Dataset
from xarray import open_dataset as _open_dataset

import xclim
from xclim import __version__ as __xclim_version__

try:
    import pytest
    from pytest_socket import SocketBlockedError
except ImportError:
    pytest = None
    SocketBlockedError = None

try:
    import pooch
except ImportError:
    warnings.warn(
        "The `pooch` library is not installed. "
        "The default cache directory for testing data will not be set."
    )
    pooch = None


logger = logging.getLogger("xclim")


__all__ = [
    "TESTDATA_BRANCH",
    "TESTDATA_CACHE_DIR",
    "TESTDATA_REPO_URL",
    "audit_url",
    "default_testdata_cache",
    "default_testdata_repo_url",
    "default_testdata_version",
    "gather_testing_data",
    "list_input_variables",
    "nimbus",
    "open_dataset",
    "populate_testing_data",
    "publish_release_notes",
    "run_doctests",
    "show_versions",
    "testing_setup_warnings",
]

default_testdata_version = "v2024.8.23"
"""Default version of the testing data to use when fetching datasets."""

default_testdata_repo_url = (
    "https://raw.githubusercontent.com/Ouranosinc/xclim-testdata/"
)
"""Default URL of the testing data repository to use when fetching datasets."""

try:
    default_testdata_cache = Path(pooch.os_cache("xclim-testdata"))
    """Default location for the testing data cache."""
except AttributeError:
    default_testdata_cache = None

TESTDATA_REPO_URL = str(os.getenv("XCLIM_TESTDATA_REPO_URL", default_testdata_repo_url))
"""Sets the URL of the testing data repository to use when fetching datasets.

Notes
-----
When running tests locally, this can be set for both `pytest` and `tox` by exporting the variable:

.. code-block:: console

    $ export XCLIM_TESTDATA_REPO_URL="https://github.com/my_username/xclim-testdata"

or setting the variable at runtime:

.. code-block:: console

    $ env XCLIM_TESTDATA_REPO_URL="https://github.com/my_username/xclim-testdata" pytest
"""

TESTDATA_BRANCH = str(os.getenv("XCLIM_TESTDATA_BRANCH", default_testdata_version))
"""Sets the branch of the testing data repository to use when fetching datasets.

Notes
-----
When running tests locally, this can be set for both `pytest` and `tox` by exporting the variable:

.. code-block:: console

    $ export XCLIM_TESTDATA_BRANCH="my_testing_branch"

or setting the variable at runtime:

.. code-block:: console

    $ env XCLIM_TESTDATA_BRANCH="my_testing_branch" pytest
"""

TESTDATA_CACHE_DIR = os.getenv("XCLIM_TESTDATA_CACHE_DIR", default_testdata_cache)
"""Sets the directory to store the testing datasets.

If not set, the default location will be used (based on ``platformdirs``, see :func:`pooch.os_cache`).

Notes
-----
When running tests locally, this can be set for both `pytest` and `tox` by exporting the variable:

.. code-block:: console

    $ export XCLIM_TESTDATA_CACHE_DIR="/path/to/my/data"

or setting the variable at runtime:

.. code-block:: console

    $ env XCLIM_TESTDATA_CACHE_DIR="/path/to/my/data" pytest
"""


def list_input_variables(
    submodules: Sequence[str] | None = None, realms: Sequence[str] | None = None
) -> dict:
    """List all possible variables names used in xclim's indicators.

    Made for development purposes. Parses all indicator parameters with the
    :py:attr:`xclim.core.utils.InputKind.VARIABLE` or `OPTIONAL_VARIABLE` kinds.

    Parameters
    ----------
    realms: Sequence of str, optional
      Restrict the output to indicators of a list of realms only. Default None, which parses all indicators.
    submodules: str, optional
      Restrict the output to indicators of a list of submodules only. Default None, which parses all indicators.

    Returns
    -------
    dict
      A mapping from variable name to indicator class.
    """
    from collections import defaultdict  # pylint: disable=import-outside-toplevel

    from xclim import indicators  # pylint: disable=import-outside-toplevel
    from xclim.core.indicator import registry  # pylint: disable=import-outside-toplevel
    from xclim.core.utils import InputKind  # pylint: disable=import-outside-toplevel

    submodules = submodules or [
        sub for sub in dir(indicators) if not sub.startswith("__")
    ]
    realms = realms or ["atmos", "ocean", "land", "seaIce"]

    variables = defaultdict(list)
    for name, ind in registry.items():
        if "." in name:
            # external submodule, submodule name is prepended to registry key
            if name.split(".")[0] not in submodules:
                continue
        elif ind.realm not in submodules:
            # official indicator : realm == submodule
            continue
        if ind.realm not in realms:
            continue

        # ok we want this one.
        for varname, meta in ind._all_parameters.items():
            if meta.kind in [
                InputKind.VARIABLE,
                InputKind.OPTIONAL_VARIABLE,
            ]:
                var = meta.default or varname
                variables[var].append(ind)

    return variables


# Publishing Tools ###


def publish_release_notes(
    style: str = "md",
    file: os.PathLike[str] | StringIO | TextIO | None = None,
    changes: str | os.PathLike[str] | None = None,
) -> str | None:
    """Format release notes in Markdown or ReStructuredText.

    Parameters
    ----------
    style : {"rst", "md"}
        Use ReStructuredText formatting or Markdown. Default: Markdown.
    file : {os.PathLike, StringIO, TextIO}, optional
        If provided, prints to the given file-like object. Otherwise, returns a string.
    changes : str or os.PathLike[str], optional
        If provided, manually points to the file where the changelog can be found.
        Assumes a relative path otherwise.

    Returns
    -------
    str, optional

    Notes
    -----
    This function is used solely for development and packaging purposes.
    """
    if isinstance(changes, str | Path):
        changes_file = Path(changes).absolute()
    else:
        changes_file = Path(__file__).absolute().parents[2].joinpath("CHANGELOG.rst")

    if not changes_file.exists():
        raise FileNotFoundError("Changelog file not found in xclim folder tree.")

    with open(changes_file, encoding="utf-8") as hf:
        changes = hf.read()

    if style == "rst":
        hyperlink_replacements = {
            r":issue:`([0-9]+)`": r"`GH/\1 <https://github.com/Ouranosinc/xclim/issues/\1>`_",
            r":pull:`([0-9]+)`": r"`PR/\1 <https://github.com/Ouranosinc/xclim/pull/\>`_",
            r":user:`([a-zA-Z0-9_.-]+)`": r"`@\1 <https://github.com/\1>`_",
        }
    elif style == "md":
        hyperlink_replacements = {
            r":issue:`([0-9]+)`": r"[GH/\1](https://github.com/Ouranosinc/xclim/issues/\1)",
            r":pull:`([0-9]+)`": r"[PR/\1](https://github.com/Ouranosinc/xclim/pull/\1)",
            r":user:`([a-zA-Z0-9_.-]+)`": r"[@\1](https://github.com/\1)",
        }
    else:
        msg = f"Formatting style not supported: {style}"
        raise NotImplementedError(msg)

    for search, replacement in hyperlink_replacements.items():
        changes = re.sub(search, replacement, changes)

    if style == "md":
        changes = changes.replace("=========\nChangelog\n=========", "# Changelog")

        titles = {r"\n(.*?)\n([\-]{1,})": "-", r"\n(.*?)\n([\^]{1,})": "^"}
        for title_expression, level in titles.items():
            found = re.findall(title_expression, changes)
            for grouping in found:
                fixed_grouping = (
                    str(grouping[0]).replace("(", r"\(").replace(")", r"\)")
                )
                search = rf"({fixed_grouping})\n([\{level}]{'{' + str(len(grouping[1])) + '}'})"
                replacement = f"{'##' if level == '-' else '###'} {grouping[0]}"
                changes = re.sub(search, replacement, changes)

        link_expressions = r"[\`]{1}([\w\s]+)\s<(.+)>`\_"
        found = re.findall(link_expressions, changes)
        for grouping in found:
            search = rf"`{grouping[0]} <.+>`\_"
            replacement = f"[{str(grouping[0]).strip()}]({grouping[1]})"
            changes = re.sub(search, replacement, changes)

    if not file:
        return changes
    if isinstance(file, Path | os.PathLike):
        with open(file, "w", encoding="utf-8") as f:
            print(changes, file=f)
    else:
        print(changes, file=file)
    return None


_xclim_deps = [
    "xclim",
    "xarray",
    "statsmodels",
    "sklearn",
    "scipy",
    "pint",
    "pandas",
    "numpy",
    "numba",
    "lmoments3",
    "jsonpickle",
    "flox",
    "dask",
    "cf_xarray",
    "cftime",
    "clisops",
    "click",
    "bottleneck",
    "boltons",
]


def show_versions(
    file: os.PathLike | StringIO | TextIO | None = None,
    deps: list[str] | None = None,
) -> str | None:
    """Print the versions of xclim and its dependencies.

    Parameters
    ----------
    file : {os.PathLike, StringIO, TextIO}, optional
        If provided, prints to the given file-like object. Otherwise, returns a string.
    deps : list of str, optional
        A list of dependencies to gather and print version information from. Otherwise, prints `xclim` dependencies.

    Returns
    -------
    str or None
    """
    dependencies: list[str]
    if deps is None:
        dependencies = _xclim_deps
    else:
        dependencies = deps

    dependency_versions = [(d, lambda mod: mod.__version__) for d in dependencies]

    deps_blob: list[tuple[str, str | None]] = []
    for modname, ver_f in dependency_versions:
        try:
            if modname in sys.modules:
                mod = sys.modules[modname]
            else:
                mod = import_module(modname)
        except (KeyError, ModuleNotFoundError):
            deps_blob.append((modname, None))
        else:
            try:
                ver = ver_f(mod)
                deps_blob.append((modname, ver))
            except AttributeError:
                deps_blob.append((modname, "installed"))

    modules_versions = "\n".join([f"{k}: {stat}" for k, stat in sorted(deps_blob)])

    installed_versions = [
        "INSTALLED VERSIONS",
        "------------------",
        f"python: {platform.python_version()}",
        f"{modules_versions}",
        f"Anaconda-based environment: {'yes' if Path(sys.base_prefix).joinpath('conda-meta').exists() else 'no'}",
    ]

    message = "\n".join(installed_versions)

    if not file:
        return message
    if isinstance(file, Path | os.PathLike):
        with open(file, "w", encoding="utf-8") as f:
            print(message, file=f)
    else:
        print(message, file=file)
    return None


# Test Data Utilities ###


def run_doctests():
    """Run the doctests for the module."""
    if pytest is None:
        raise ImportError(
            "The `pytest` package is required to run the doctests. "
            "You can install it with `pip install pytest` or `pip install xclim[dev]`."
        )

    cmd = [
        f"--rootdir={Path(__file__).absolute().parent}",
        "--numprocesses=0",
        "--xdoctest",
        f"{Path(__file__).absolute().parents[1]}",
    ]

    sys.exit(pytest.main(cmd))


def testing_setup_warnings():
    """Warn users about potential incompatibilities between xclim and xclim-testdata versions."""
    if (
        re.match(r"^\d+\.\d+\.\d+$", __xclim_version__)
        and TESTDATA_BRANCH != default_testdata_version
    ):
        # This does not need to be emitted on GitHub Workflows and ReadTheDocs
        if not os.getenv("CI") and not os.getenv("READTHEDOCS"):
            warnings.warn(
                f"`xclim` stable ({__xclim_version__}) is running tests against a non-default branch of the testing data. "
                "It is possible that changes to the testing data may be incompatible with some assertions in this version. "
                f"Please be sure to check {TESTDATA_REPO_URL} for more information.",
            )

    if re.match(r"^v\d+\.\d+\.\d+", TESTDATA_BRANCH):
        # Find the date of last modification of xclim source files to generate a calendar version
        install_date = dt.strptime(
            time.ctime(os.path.getmtime(xclim.__file__)),
            "%a %b %d %H:%M:%S %Y",
        )
        install_calendar_version = (
            f"{install_date.year}.{install_date.month}.{install_date.day}"
        )

        if Version(TESTDATA_BRANCH) > Version(install_calendar_version):
            warnings.warn(
                f"The installation date of `xclim` ({install_date.ctime()}) "
                f"predates the last release of testing data ({TESTDATA_BRANCH}). "
                "It is very likely that the testing data is incompatible with this build of `xclim`.",
            )


def load_registry(
    branch: str = TESTDATA_BRANCH, repo: str = TESTDATA_REPO_URL
) -> dict[str, str]:
    """Load the registry file for the test data.

    Returns
    -------
    dict
        Dictionary of filenames and hashes.
    """
    remote_registry = audit_url(f"{repo}/{branch}/data/registry.txt")

    if branch != default_testdata_version:
        custom_registry_folder = Path(
            str(ilr.files("xclim").joinpath(f"testing/{branch}"))
        )
        custom_registry_folder.mkdir(parents=True, exist_ok=True)
        registry_file = custom_registry_folder.joinpath("registry.txt")
        urlretrieve(remote_registry, registry_file)  # noqa: S310

    elif repo != default_testdata_repo_url:
        registry_file = Path(str(ilr.files("xclim").joinpath("testing/registry.txt")))
        urlretrieve(remote_registry, registry_file)  # noqa: S310

    registry_file = Path(str(ilr.files("xclim").joinpath("testing/registry.txt")))
    if not registry_file.exists():
        raise FileNotFoundError(f"Registry file not found: {registry_file}")

    # Load the registry file
    with registry_file.open(encoding="utf-8") as f:
        registry = {line.split()[0]: line.split()[1] for line in f}
    return registry


def nimbus(  # noqa: PR01
    repo: str = TESTDATA_REPO_URL,
    branch: str = TESTDATA_BRANCH,
    cache_dir: str | Path = TESTDATA_CACHE_DIR,
    data_updates: bool = True,
):
    """Pooch registry instance for xclim test data.

    Parameters
    ----------
    repo : str
        URL of the repository to use when fetching testing datasets.
    branch : str
        Branch of repository to use when fetching testing datasets.
    cache_dir : str or Path
        The path to the directory where the data files are stored.
    data_updates : bool
        If True, allow updates to the data files. Default is True.

    Returns
    -------
    pooch.Pooch
        The Pooch instance for accessing the xclim testing data.

    Notes
    -----
    There are three environment variables that can be used to control the behaviour of this registry:
        - ``XCLIM_TESTDATA_CACHE_DIR``: If this environment variable is set, it will be used as the base directory to
          store the data files. The directory should be an absolute path (i.e., it should start with ``/``).
          Otherwise,the default location will be used (based on ``platformdirs``, see :py:func:`pooch.os_cache`).
        - ``XCLIM_TESTDATA_REPO_URL``: If this environment variable is set, it will be used as the URL of the repository
          to use when fetching datasets. Otherwise, the default repository will be used.
        - ``XCLIM_TESTDATA_BRANCH``: If this environment variable is set, it will be used as the branch of the repository
          to use when fetching datasets. Otherwise, the default branch will be used.

    Examples
    --------
    Using the registry to download a file:

    .. code-block:: python

        import xarray as xr
        from xclim.testing.helpers import nimbus

        example_file = nimbus().fetch("example.nc")
        data = xr.open_dataset(example_file)
    """
    if pooch is None:
        raise ImportError(
            "The `pooch` package is required to fetch the xclim testing data. "
            "You can install it with `pip install pooch` or `pip install xclim[dev]`."
        )

    remote = audit_url(f"{repo}/{branch}/data")
    return pooch.create(
        path=cache_dir,
        base_url=remote,
        version=default_testdata_version,
        version_dev=branch,
        allow_updates=data_updates,
        registry=load_registry(branch=branch, repo=repo),
    )


# idea copied from raven that it borrowed from xclim that borrowed it from xarray that was borrowed from Seaborn
def open_dataset(
    name: str | os.PathLike[str],
    dap_url: str | None = None,
    branch: str = TESTDATA_BRANCH,
    repo: str = TESTDATA_REPO_URL,
    cache_dir: str | os.PathLike[str] | None = TESTDATA_CACHE_DIR,
    **kwargs,
) -> Dataset:
    r"""Open a dataset from the online GitHub-like repository.

    If a local copy is found then always use that to avoid network traffic.

    Parameters
    ----------
    name : str
        Name of the file containing the dataset.
    dap_url : str, optional
        URL to OPeNDAP folder where the data is stored. If supplied, supersedes github_url.
    branch : str
        Branch of the repository to use when fetching datasets.
    repo: str
        URL of the repository to use when fetching testing datasets.
    cache_dir : Path
        The directory in which to search for and write cached data.
    \*\*kwargs
        For NetCDF files, keywords passed to :py:func:`xarray.open_dataset`.

    Returns
    -------
    Union[Dataset, Path]

    Raises
    ------
    OSError
        If the file is not found in the cache directory or cannot be read.

    See Also
    --------
    xarray.open_dataset
    """
    if cache_dir is None:
        raise ValueError(
            "The cache directory must be set. "
            "Please set the `cache_dir` parameter or the `XCLIM_DATA_DIR` environment variable."
        )

    if dap_url:
        dap_target = urljoin(dap_url, str(name))
        try:
            return _open_dataset(audit_url(dap_target, context="OPeNDAP"), **kwargs)
        except URLError:
            raise
        except OSError as err:
            msg = f"OPeNDAP file not read. Verify that the service is available: {dap_target}"
            raise OSError(msg) from err

    local_file = Path(cache_dir).joinpath(name)
    if not local_file.exists():
        try:
            local_file = nimbus(branch=branch, repo=repo, cache_dir=cache_dir).fetch(
                name
            )
        except OSError as err:
            msg = f"File not found locally. Verify that the testing data is available in remote: {local_file}"
            raise OSError(msg) from err
    try:
        ds = _open_dataset(local_file, **kwargs)
        return ds
    except OSError:
        raise


def populate_testing_data(
    temp_folder: Path | None = None,
    repo: str = TESTDATA_REPO_URL,
    branch: str = TESTDATA_BRANCH,
    local_cache: Path = TESTDATA_CACHE_DIR,
) -> None:
    """Populate the local cache with the testing data.

    Parameters
    ----------
    temp_folder : Path, optional
        Path to a temporary folder to use as the local cache. If not provided, the default location will be used.
    repo : str, optional
        URL of the repository to use when fetching testing datasets.
    branch : str, optional
        Branch of xclim-testdata to use when fetching testing datasets.
    local_cache : Path
        The path to the local cache. Defaults to the location set by the platformdirs library.
        The testing data will be downloaded to this local cache.

    Returns
    -------
    None
    """
    # Create the Pooch instance
    n = nimbus(repo=repo, branch=branch, cache_dir=temp_folder or local_cache)

    # Download the files
    errored_files = []
    for file in load_registry():
        try:
            n.fetch(file)
        except HTTPError:
            msg = f"File `{file}` not accessible in remote repository."
            logging.error(msg)
            errored_files.append(file)
        except SocketBlockedError as err:  # noqa
            msg = (
                "Unable to access registry file online. Testing suite is being run with `--disable-socket`. "
                "If you intend to run tests with this option enabled, please download the file beforehand with the "
                "following console command: `$ xclim prefetch_testing_data`."
            )
            raise SocketBlockedError(msg) from err
        else:
            logging.info("Files were downloaded successfully.")

    if errored_files:
        logging.error(
            "The following files were unable to be downloaded: %s",
            errored_files,
        )


def gather_testing_data(
    worker_cache_dir: str | os.PathLike[str] | Path,
    worker_id: str,
    _cache_dir: str | os.PathLike[str] | None = TESTDATA_CACHE_DIR,
):
    """Gather testing data across workers."""
    if _cache_dir is None:
        raise ValueError(
            "The cache directory must be set. "
            "Please set the `cache_dir` parameter or the `XCLIM_DATA_DIR` environment variable."
        )
    cache_dir = Path(_cache_dir)

    if worker_id == "master":
        populate_testing_data(branch=TESTDATA_BRANCH)
    else:
        if platform.system() == "Windows":
            if not cache_dir.joinpath(default_testdata_version).exists():
                raise FileNotFoundError(
                    "Testing data not found and UNIX-style file-locking is not supported on Windows. "
                    "Consider running `$ xclim prefetch_testing_data` to download testing data beforehand."
                )
        else:
            cache_dir.mkdir(exist_ok=True, parents=True)
            lockfile = cache_dir.joinpath(".lock")
            test_data_being_written = FileLock(lockfile)
            with test_data_being_written:
                # This flag prevents multiple calls from re-attempting to download testing data in the same pytest run
                populate_testing_data(branch=TESTDATA_BRANCH)
                cache_dir.joinpath(".data_written").touch()
            with test_data_being_written.acquire():
                if lockfile.exists():
                    lockfile.unlink()
        copytree(cache_dir.joinpath(default_testdata_version), worker_cache_dir)


# Testing Utilities ###


def audit_url(url: str, context: str | None = None) -> str:
    """Check if the URL is well-formed.

    Raises
    ------
    URLError
        If the URL is not well-formed.
    """
    msg = ""
    result = urlparse(url)
    if result.scheme == "http":
        msg = f"{context if context else ''} URL is not using secure HTTP: '{url}'".strip()
    if not all([result.scheme, result.netloc]):
        msg = f"{context if context else ''} URL is not well-formed: '{url}'".strip()

    if msg:
        logger.error(msg)
        raise URLError(msg)
    return url



from __future__ import annotations


def rstjinja(app, docname, source):
    """Render our pages as a jinja template for fancy templating goodness."""
    # Make sure we're outputting HTML
    if app.builder.format != "html":
        return
    src = source[0]
    rendered = app.builder.templates.render_string(src, app.config.html_context)
    source[0] = rendered


def setup(app):
    app.connect("source-read", rstjinja)
